import collections
import dataclasses
import fractions
import itertools
import logging
import queue
import time
import traceback
from typing import Optional, Tuple

from .. import conversions
from .ephemera import (
    CallbackCommand,
    CallbackEvent,
    ChangeCommand,
    ChangeEvent,
    ClockContext,
    ClockState,
    EventType,
    Moment,
    TimeUnit,
)
from .eventqueue import EventQueue

logger = logging.getLogger("supriya.clocks")


class BaseClock:

    ### CLASS VARIABLES ###

    _valid_quantizations = frozenset(
        [
            "8M",
            "4M",
            "2M",
            "1M",
            "1/2",
            "1/2T",
            "1/4",
            "1/4T",
            "1/8",
            "1/8T",
            "1/16",
            "1/16T",
            "1/32",
            "1/32T",
            "1/64",
            "1/64T",
            "1/128",
        ]
    )

    ### INITIALIZER ###

    def __init__(self):
        self._name = None
        self._counter = itertools.count()
        self._command_deque = collections.deque()
        self._event_queue = EventQueue()
        self._is_running = False
        self._slop = 0.001
        self._events_by_id = {}
        self._measure_relative_event_ids = set()
        self._offset_relative_event_ids = set()
        self._state = ClockState(
            beats_per_minute=120.0,
            initial_seconds=0.0,
            previous_measure=1,
            previous_offset=0.0,
            previous_seconds=0.0,
            previous_time_signature_change_offset=0.0,
            time_signature=(4, 4),
        )

    ### TIME METHODS ###

    def _get_cue_point(self, seconds, quantization):
        moment = self._seconds_to_moment(seconds)
        if quantization is None:
            offset, measure = moment.offset, None
        elif "M" in quantization:
            measure_grid = int(quantization[0])
            grid_offset = moment.measure - 1 + (moment.measure_offset > 0)
            count, modulus = divmod(grid_offset, measure_grid)
            count += modulus > 0
            measure = (count * measure_grid) + 1
            offset = self._measure_to_offset(measure)
        else:
            measure = None
            fraction_grid = self.quantization_to_beats(quantization)
            div, mod = divmod(moment.offset, fraction_grid)
            offset = float(div * fraction_grid)
            if mod:
                offset += fraction_grid
        seconds = self._offset_to_seconds(offset)
        logger.debug(
            f"[{self.name}] ... ... Cueing {quantization} to "
            f"{seconds}:s / {offset}:o / {measure}:m"
        )
        return seconds, offset, measure

    def _get_schedule_point(self, schedule_at: float, time_unit: TimeUnit):
        measure: Optional[int] = None
        offset: Optional[float] = None
        if time_unit == TimeUnit.MEASURES:
            measure = int(schedule_at)
            offset = self._measure_to_offset(measure)
            seconds: float = self._offset_to_seconds(offset)
        elif time_unit == TimeUnit.BEATS:
            offset = float(schedule_at)
            seconds = self._offset_to_seconds(schedule_at)
        else:
            seconds = float(schedule_at)
        return seconds, offset, measure

    def _measure_to_offset(self, measure: int) -> float:
        return conversions.measure_to_offset(
            measure,
            self._state.time_signature,
            self._state.previous_measure,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_measure(self, offset: float) -> int:
        return conversions.offset_to_measure(
            offset,
            self._state.time_signature,
            self._state.previous_measure,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_measure_offset(self, offset: float) -> float:
        return conversions.offset_to_measure_offset(
            offset,
            self._state.time_signature,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_moment(self, offset: float) -> Moment:
        seconds = self._offset_to_seconds(offset)
        measure, measure_offset = divmod(
            offset - self._state.previous_time_signature_change_offset,
            self._state.time_signature[0] / self._state.time_signature[1],
        )
        return Moment(
            beats_per_minute=self._state.beats_per_minute,
            measure=int(measure + self._state.previous_measure),
            measure_offset=measure_offset,
            offset=offset,
            seconds=seconds,
            time_signature=self._state.time_signature,
        )

    def _offset_to_seconds(self, offset: float) -> float:
        return conversions.offset_to_seconds(
            beats_per_minute=self._state.beats_per_minute,
            current_offset=offset,
            previous_offset=self._state.previous_offset,
            previous_seconds=self._state.previous_seconds,
            beat_duration=1 / self._state.time_signature[1],
        )

    def _seconds_to_moment(self, seconds):
        offset = self._seconds_to_offset(seconds)
        measure, measure_offset = divmod(
            offset - self._state.previous_time_signature_change_offset,
            self._state.time_signature[0] / self._state.time_signature[1],
        )
        return Moment(
            beats_per_minute=self._state.beats_per_minute,
            measure=int(measure + self._state.previous_measure),
            measure_offset=measure_offset,
            offset=offset,
            seconds=seconds,
            time_signature=self._state.time_signature,
        )

    def _seconds_to_offset(self, seconds: float) -> float:
        return conversions.seconds_to_offset(
            beats_per_minute=self._state.beats_per_minute,
            current_time=seconds,
            previous_offset=self._state.previous_offset,
            previous_seconds=self._state.previous_seconds,
            beat_duration=1 / self._state.time_signature[1],
        )

    ### SCHEDULING METHODS ###

    def _cancel(self, event_id) -> Optional[Tuple]:
        event = self._events_by_id.pop(event_id, None)
        if event is not None and not isinstance(
            event, (CallbackCommand, ChangeCommand)
        ):
            self._event_queue.remove(event)
            if event.offset is not None:
                self._offset_relative_event_ids.remove(event.event_id)
                if event.measure is not None:
                    self._measure_relative_event_ids.remove(event.event_id)
        return event

    def _enqueue_command(self, command):
        self._events_by_id[command.event_id] = command
        self._command_deque.append(command)
        if isinstance(command, CallbackCommand):
            logger.debug(
                f"[{self.name}] Enqueued {type(command).__name__} ({command.event_id}) {command.procedure}"
            )
        else:
            logger.debug(
                f"[{self.name}] Enqueued {type(command).__name__} ({command.event_id})"
            )

    def _enqueue_event(self, event):
        self._events_by_id[event.event_id] = event
        self._event_queue.put(event)
        if event.offset is not None:
            self._offset_relative_event_ids.add(event.event_id)
            if event.measure is not None:
                self._measure_relative_event_ids.add(event.event_id)

    def _process_perform_event_loop(self, current_moment):
        # There may be items in the queue which have been flagged "removed".
        # They contribute to the queue's size, but cannot be retrieved by .get().
        try:
            event = self._event_queue.get()
        except queue.Empty:
            return None, None, True, False
        if self._events_by_id.pop(event.event_id, None) is None:
            return None, None, True, False
        if current_moment.seconds < event.seconds:
            self._enqueue_event(event)
            return None, None, False, True
        if event.offset is not None:
            desired_moment = self._offset_to_moment(event.offset)
        else:
            desired_moment = self._seconds_to_moment(event.seconds)
        if event.offset is not None and event.offset != desired_moment.offset:
            raise RuntimeError(
                f"Offset mismatch: {event.offset} vs {desired_moment.offset}"
            )
        return event, desired_moment, False, False

    def _perform_callback_event(self, event, current_moment, desired_moment):
        logger.debug(
            f"[{self.name}] ... ... Performing {event.procedure} at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s / "
            f"{desired_moment.offset}:o"
        )
        context = ClockContext(current_moment, desired_moment, event)
        args = event.args or ()
        kwargs = event.kwargs or {}
        try:
            result = event.procedure(context, *args, **kwargs)
        except Exception:
            traceback.print_exc()
            return
        self._process_callback_event_result(desired_moment, event, result)

    def _process_callback_event_result(self, desired_moment, event, result):
        try:
            delta, unit = result
        except TypeError:
            delta, unit = result, TimeUnit.BEATS
        if delta is None or delta <= 0:
            return
        kwargs = {"invocations": event.invocations + 1, "measure": None, "offset": None}
        if unit == TimeUnit.MEASURES:
            kwargs["measure"] = desired_moment.measure + delta
            kwargs["offset"] = self._measure_to_offset(kwargs["measure"])
        if unit == TimeUnit.BEATS:
            kwargs["offset"] = desired_moment.offset + delta
        if unit in (TimeUnit.BEATS, TimeUnit.MEASURES):
            kwargs["seconds"] = self._offset_to_seconds(kwargs["offset"])
        if unit == TimeUnit.SECONDS:
            kwargs["seconds"] = desired_moment.seconds + delta
        logger.debug(
            f"[{self.name}] ... ... ... Rescheduling "
            f"{event.procedure} ({event.event_id}) at {kwargs['seconds'] - self._state.initial_seconds}s"
        )
        event = event._replace(**kwargs)
        self._enqueue_event(event)

    def _perform_change_event(self, event, current_moment, desired_moment):
        logger.debug(
            f"[{self.name}] ... ... Changing at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s"
        )
        # TODO: current offset is misleading here
        if event.time_signature is not None:
            new_duration = event.time_signature[0] / event.time_signature[1]
            if desired_moment.measure_offset < new_duration:
                # On the downbeat
                self._state = self._state._replace(
                    previous_measure=desired_moment.measure,
                    previous_time_signature_change_offset=(
                        desired_moment.offset - desired_moment.measure_offset
                    ),
                    time_signature=event.time_signature,
                )
            else:
                # Moving from a longer time signature to a shorter one
                # Advance to the next downbeat immediately
                self._state = self._state._replace(
                    previous_measure=desired_moment.measure + 1,
                    previous_time_signature_change_offset=desired_moment.offset,
                    time_signature=event.time_signature,
                )
            self._reschedule_measure_relative_events()
            current_moment = dataclasses.replace(
                current_moment,
                time_signature=self._state.time_signature,
                measure=self._offset_to_measure(current_moment.offset),
                measure_offset=self._offset_to_measure_offset(current_moment.offset),
            )
        if event.beats_per_minute is not None:
            # If the tempo changed, we need to revise our offset math
            self._state = self._state._replace(
                beats_per_minute=float(event.beats_per_minute),
                previous_seconds=desired_moment.seconds,
                previous_offset=desired_moment.offset,
            )
            self._reschedule_offset_relative_events()
            new_current_offset = self._seconds_to_offset(current_moment.seconds)
            logger.debug(
                f"[{self.name}] ... ... ... Revised offset from "
                f"{current_moment.offset} to {new_current_offset}"
            )
            current_moment = dataclasses.replace(
                current_moment, offset=new_current_offset
            )
            return current_moment, False
            # if new_current_offset < current_moment.offset:
            #    return current_moment, False
        return current_moment, True

    def _perform_events(self, current_moment: Moment):
        logger.debug(
            f"[{self.name}] ... Ready to perform at "
            f"{current_moment.seconds - self._state.initial_seconds}:s / "
            f"{current_moment.offset}:o"
        )
        while self._is_running and self._event_queue.qsize():
            (
                event,
                desired_moment,
                should_continue,
                should_break,
            ) = self._process_perform_event_loop(current_moment)
            if should_continue:
                continue
            elif should_break:
                break
            if event.event_type == EventType.CHANGE:
                current_moment, should_continue = self._perform_change_event(
                    event, current_moment, desired_moment
                )
                if not should_continue:
                    break
            else:
                self._perform_callback_event(event, current_moment, desired_moment)
                self._process_command_deque()
        return current_moment

    def _process_command_deque(self, first_run=False):
        while self._command_deque:
            logger.debug(f"[{self.name}] ... Processing command deque ({first_run})")
            command = self._command_deque.popleft()
            if self._events_by_id.pop(command.event_id, None) is None:
                continue
            schedule_at = command.schedule_at
            logger.debug(f"[{self.name}] ... ... Scheduled at {schedule_at}")
            if command.quantization is not None:
                # If the command was queued before the clock was started,
                # reset its reference time
                if first_run and schedule_at < self._state.initial_seconds:
                    schedule_at = self._state.initial_seconds
                seconds, offset, measure = self._get_cue_point(
                    schedule_at, command.quantization
                )
            elif command.time_unit in (TimeUnit.BEATS, TimeUnit.MEASURES):
                seconds, offset, measure = self._get_schedule_point(
                    schedule_at, command.time_unit
                )
            else:
                if first_run:
                    schedule_at += self._state.initial_seconds
                seconds, offset, measure = schedule_at, None, None
            if isinstance(command, CallbackCommand):
                event = CallbackEvent(
                    args=command.args,
                    event_id=command.event_id,
                    kwargs=command.kwargs,
                    measure=measure,
                    offset=offset,
                    procedure=command.procedure,
                    seconds=seconds,
                    event_type=command.event_type,
                    invocations=0,
                )
            else:
                event = ChangeEvent(
                    beats_per_minute=command.beats_per_minute,
                    event_id=command.event_id,
                    event_type=command.event_type,
                    measure=measure,
                    offset=offset,
                    seconds=seconds,
                    time_signature=command.time_signature,
                )
            self._enqueue_event(event)
            logger.debug(
                f"[{self.name}] ... ... Enqueued {type(event).__name__} "
                f"({event.event_id}) for {event.seconds}:s / {event.offset}:o"
            )

    def _reschedule_offset_relative_events(self):
        for event_id in tuple(self._offset_relative_event_ids):
            event = self._cancel(event_id)
            if event is None:
                self._offset_relative_event_ids.remove(event_id)
                continue
            seconds = self._offset_to_seconds(event.offset)
            logger.debug(
                f"[{self.name}] ... ... ... Rescheduling offset-relative event "
                f"({event.event_id}) from "
                f"{event.seconds - self._state.initial_seconds}:s to "
                f"{seconds - self._state.initial_seconds}:s"
            )
            self._enqueue_event(event._replace(seconds=seconds))

    def _reschedule_measure_relative_events(self):
        for event_id in tuple(self._measure_relative_event_ids):
            event = self._cancel(event_id)
            if event is None:
                self._measure_relative_event_ids.remove(event_id)
                continue
            offset = self._measure_to_offset(event.measure)
            seconds = self._offset_to_seconds(offset)
            logger.debug(
                f"[{self.name}] ... ... ... Rescheduling measure-relative event from "
                f"offset {event.offset} to {offset}"
            )
            self._enqueue_event(event._replace(offset=offset, seconds=seconds))

    def _start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ):
        if self._is_running:
            raise RuntimeError("Already started")
        if initial_time is None:
            initial_time = self.get_current_time()
        self._state = ClockState(
            beats_per_minute=beats_per_minute or self._state.beats_per_minute,
            initial_seconds=initial_time,
            previous_measure=int(initial_measure),
            previous_offset=float(initial_offset),
            previous_seconds=float(initial_time),
            previous_time_signature_change_offset=float(initial_offset),
            time_signature=time_signature or self._state.time_signature,
        )
        self._is_running = True

    def _stop(self):
        if not self._is_running:
            return False
        self._is_running = False
        return True

    ### PUBLIC METHODS ###

    def cancel(self, event_id) -> Optional[Tuple]:
        logger.debug(f"[{self.name}] Canceling {event_id}")
        event = self._cancel(event_id)
        return event

    def change(
        self,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> Optional[int]:
        if not self._is_running:
            self._state = self._state._replace(
                beats_per_minute=beats_per_minute or self._state.beats_per_minute,
                time_signature=time_signature or self._state.time_signature,
            )
            return None
        event_id = next(self._counter)
        command = ChangeCommand(
            event_id=event_id,
            event_type=EventType.CHANGE,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
            quantization=None,
            schedule_at=self.get_current_time(),
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def cue(
        self,
        procedure,
        *,
        args=None,
        event_type: EventType = EventType.SCHEDULE,
        kwargs=None,
        quantization: str = None,
    ) -> int:
        if event_type <= 0:
            raise ValueError(f"Invalid event type {event_type}")
        elif quantization is not None and quantization not in self._valid_quantizations:
            raise ValueError(f"Invalid quantization: {quantization}")
        event_id = next(self._counter)
        command = CallbackCommand(
            args=args,
            event_id=event_id,
            event_type=event_type,
            kwargs=kwargs,
            procedure=procedure,
            quantization=quantization,
            schedule_at=self.get_current_time() if self.is_running else 0,
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def cue_change(
        self,
        *,
        beats_per_minute: Optional[float] = None,
        quantization: str = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> int:
        if quantization is not None and quantization not in self._valid_quantizations:
            raise ValueError(f"Invalid quantization: {quantization}")
        event_id = next(self._counter)
        command = ChangeCommand(
            beats_per_minute=beats_per_minute,
            event_id=event_id,
            event_type=EventType.CHANGE,
            quantization=quantization,
            schedule_at=self.get_current_time() if self.is_running else 0,
            time_signature=time_signature,
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def get_current_time(self) -> float:
        return time.time()

    def peek(self):
        try:
            return self._event_queue.peek()
        except queue.Empty:
            pass

    @classmethod
    def quantization_to_beats(cls, quantization):
        fraction = fractions.Fraction(quantization.replace("T", ""))
        if "T" in quantization:
            fraction *= fractions.Fraction(2, 3)
        return float(fraction)

    def reschedule(
        self, event_id, *, schedule_at=0.0, time_unit=TimeUnit.BEATS
    ) -> Optional[int]:
        event_or_command = self.cancel(event_id)
        if event_or_command is None:
            return None
        if isinstance(event_or_command, (CallbackCommand, ChangeCommand)):
            command = event_or_command._replace(
                schedule_at=schedule_at, time_unit=time_unit
            )
        elif isinstance(event_or_command, CallbackEvent):
            command = CallbackCommand(
                args=event_or_command.args,
                event_id=event_or_command.event_id,
                event_type=event_or_command.event_type,
                kwargs=event_or_command.kwargs,
                procedure=event_or_command.procedure,
                quantization=None,
                schedule_at=schedule_at,
                time_unit=time_unit,
            )
        elif isinstance(event_or_command, ChangeEvent):
            command = ChangeCommand(
                beats_per_minute=event_or_command.beats_per_minute,
                event_id=event_or_command.event_id,
                event_type=EventType.CHANGE,
                quantization=None,
                schedule_at=schedule_at,
                time_signature=event_or_command.time_signature,
                time_unit=time_unit,
            )
        else:
            raise ValueError(event_or_command)
        self._enqueue_command(command)
        return event_id

    def schedule(
        self,
        procedure,
        *,
        event_type: EventType = EventType.SCHEDULE,
        schedule_at: float = 0.0,
        time_unit: TimeUnit = TimeUnit.BEATS,
        args=None,
        kwargs=None,
    ) -> int:
        logger.debug(f"[{self.name}] Scheduling {procedure}")
        if event_type <= 0:
            raise ValueError(f"Invalid event type {event_type}")
        event_id = next(self._counter)
        command = CallbackCommand(
            args=args,
            event_id=event_id,
            event_type=event_type,
            kwargs=kwargs,
            procedure=procedure,
            quantization=None,
            schedule_at=schedule_at,
            time_unit=time_unit,
        )
        self._enqueue_command(command)
        return event_id

    def schedule_change(
        self,
        *,
        beats_per_minute: Optional[float] = None,
        schedule_at: float = 0.0,
        time_signature: Optional[Tuple[int, int]] = None,
        time_unit: TimeUnit = TimeUnit.BEATS,
    ) -> int:
        event_id = next(self._counter)
        command = ChangeCommand(
            beats_per_minute=beats_per_minute,
            event_id=event_id,
            event_type=EventType.CHANGE,
            quantization=None,
            schedule_at=schedule_at,
            time_signature=time_signature,
            time_unit=time_unit,
        )
        self._enqueue_command(command)
        return event_id

    ### PUBLIC PROPERTIES ###

    @property
    def beats_per_minute(self):
        return self._state.beats_per_minute

    @property
    def is_running(self):
        return self._is_running

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def slop(self) -> float:
        return self._slop

    @slop.setter
    def slop(self, slop: float):
        if slop <= 0:
            raise ValueError(slop)
        self._slop = float(slop)

    @property
    def time_signature(self):
        return self._state.time_signature



from supriya import EnvelopeShape, utils
from supriya.system import SupriyaValueObject

from .mixins import OutputProxy


class Envelope(SupriyaValueObject):
    """
    An envelope.

    ::

        >>> envelope = supriya.synthdefs.Envelope()
        >>> envelope
        Envelope()

    ::

        >>> envelope.serialize()
        [0.0, 2.0, -99.0, -99.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0]

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_envelope_segments",
        "_initial_amplitude",
        "_loop_node",
        "_offset",
        "_release_node",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        amplitudes=(0, 1, 0),
        durations=(1, 1),
        curves=("linear", "linear"),
        release_node=None,
        loop_node=None,
        offset=None,
    ):
        assert len(amplitudes)
        assert len(durations) and len(durations) == (len(amplitudes) - 1)
        if isinstance(curves, (int, float, str, EnvelopeShape, OutputProxy)):
            curves = [curves]
        elif curves is None:
            curves = []
        self._release_node = release_node
        self._loop_node = loop_node
        self._offset = offset
        self._initial_amplitude = amplitudes[0]
        self._envelope_segments = tuple(
            utils.zip_sequences(amplitudes[1:], durations, curves)
        )

    ### PUBLIC METHODS ###

    @classmethod
    def adsr(
        cls,
        attack_time=0.01,
        decay_time=0.3,
        sustain=0.5,
        release_time=1.0,
        peak=1.0,
        curve=-4.0,
        bias=0.0,
    ):
        amplitudes = [x + bias for x in [0, peak, peak * sustain, 0]]
        durations = [attack_time, decay_time, release_time]
        curves = [curve]
        release_node = 2
        return Envelope(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
        )

    @classmethod
    def asr(cls, attack_time=0.01, sustain=1.0, release_time=1.0, curve=-4.0):
        amplitudes = [0, sustain, 0]
        durations = [attack_time, release_time]
        curves = [curve]
        release_node = 1
        return Envelope(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
        )

    @classmethod
    def from_segments(
        cls,
        initial_amplitude=0,
        segments=None,
        release_node=None,
        loop_node=None,
        offset=None,
    ):
        amplitudes = [initial_amplitude]
        durations = []
        curves = []
        for amplitude, duration, curve in segments:
            amplitudes.append(amplitude)
            durations.append(duration)
            curves.append(curve)
        return cls(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
            loop_node=loop_node,
            offset=offset,
        )

    @classmethod
    def percussive(cls, attack_time=0.01, release_time=1.0, amplitude=1.0, curve=-4.0):
        """
        Make a percussion envelope.

        ::

            >>> import supriya.synthdefs
            >>> envelope = supriya.synthdefs.Envelope.percussive()
            >>> envelope
            Envelope(
                curves=(-4.0, -4.0),
                durations=(0.01, 1.0),
            )

        ::

            >>> envelope.serialize()
            [0.0, 2.0, -99.0, -99.0, 1.0, 0.01, 5.0, -4.0, 0.0, 1.0, 5.0, -4.0]

        """
        amplitudes = [0, amplitude, 0]
        durations = [attack_time, release_time]
        curves = [curve]
        return Envelope(amplitudes=amplitudes, durations=durations, curves=curves)

    @classmethod
    def linen(
        cls, attack_time=0.01, sustain_time=1.0, release_time=1.0, level=1.0, curve=1
    ):
        amplitudes = [0, level, level, 0]
        durations = [attack_time, sustain_time, release_time]
        curves = [curve]
        return Envelope(amplitudes=amplitudes, durations=durations, curves=curves)

    def serialize(self, for_interpolation=False):
        result = []
        if for_interpolation:
            result.append(self.offset or 0)
            result.append(self.initial_amplitude)
            result.append(len(self.envelope_segments))
            result.append(self.duration)
            for amplitude, duration, curve in self._envelope_segments:
                result.append(duration)
                if isinstance(curve, str):
                    shape = EnvelopeShape.from_expr(curve)
                    shape = int(shape)
                    curve = 0.0
                else:
                    shape = 5
                result.append(shape)
                result.append(curve)
                result.append(amplitude)
        else:
            result.append(self.initial_amplitude)
            result.append(len(self.envelope_segments))
            release_node = self.release_node
            if release_node is None:
                release_node = -99
            result.append(release_node)
            loop_node = self.loop_node
            if loop_node is None:
                loop_node = -99
            result.append(loop_node)
            for amplitude, duration, curve in self._envelope_segments:
                result.append(amplitude)
                result.append(duration)
                if isinstance(curve, str):
                    shape = EnvelopeShape.from_expr(curve)
                    shape = int(shape)
                    curve = 0.0
                else:
                    shape = 5
                result.append(shape)
                result.append(curve)
        serialized = []
        for x in result:
            if isinstance(x, int):
                x = float(x)
            serialized.append(x)
        return serialized

    @classmethod
    def triangle(cls, duration=1.0, amplitude=1.0):
        """
        Make a triangle envelope.

        ::

            >>> import supriya.synthdefs
            >>> envelope = supriya.synthdefs.Envelope.triangle()
            >>> envelope
            Envelope(
                durations=(0.5, 0.5),
            )

        ::

            >>> envelope.serialize()
            [0.0, 2.0, -99.0, -99.0, 1.0, 0.5, 1.0, 0.0, 0.0, 0.5, 1.0, 0.0]

        """
        amplitudes = [0, amplitude, 0]
        duration = duration / 2.0
        durations = [duration, duration]
        return Envelope(amplitudes=amplitudes, durations=durations)

    ### PUBLIC PROPERTIES ###

    @property
    def amplitudes(self):
        return (self.initial_amplitude,) + tuple(_[0] for _ in self.envelope_segments)

    @property
    def curves(self):
        return tuple(_[2] for _ in self.envelope_segments)

    @property
    def duration(self):
        return sum(self.durations)

    @property
    def durations(self):
        return tuple(_[1] for _ in self.envelope_segments)

    @property
    def envelope_segments(self):
        return self._envelope_segments

    @property
    def initial_amplitude(self):
        return self._initial_amplitude

    @property
    def loop_node(self):
        return self._loop_node

    @property
    def offset(self):
        return self._offset

    @property
    def release_node(self):
        return self._release_node



from typing import Type
from uuid import uuid4

from uqbar.objects import new

from .events import Event, NoteEvent
from .patterns import Pattern, SequencePattern


class EventPattern(Pattern):
    """
    Akin to SuperCollider's Pbind.
    """

    def __init__(self, event_type: Type[Event] = NoteEvent, **patterns):
        self._patterns = patterns
        self._event_type = event_type

    def _iterate(self, state=None):
        patterns = self._prepare_patterns()
        iterator_pairs = sorted(patterns.items())
        while True:
            event = {}
            for key, pattern in iterator_pairs:
                try:
                    event[key] = next(pattern)
                except StopIteration:
                    return
            if (yield self.event_type(uuid4(), **event)):
                return

    def _prepare_patterns(self):
        patterns = self._patterns.copy()
        for name, pattern in sorted(patterns.items()):
            if not isinstance(pattern, Pattern):
                pattern = SequencePattern([pattern], iterations=None)
            patterns[name] = iter(pattern)
        return patterns

    @property
    def event_type(self) -> Type[Event]:
        return self._event_type

    @property
    def is_infinite(self):
        for value in self._patterns.values():
            if isinstance(value, Pattern) and not value.is_infinite:
                return False
        return True


class MonoEventPattern(EventPattern):
    """
    Akin to SuperCollider's Pmono.
    """

    def _iterate(self, state=None):
        id_ = uuid4()
        patterns = self._prepare_patterns()
        iterator_pairs = sorted(patterns.items())
        while True:
            event = {}
            for key, pattern in iterator_pairs:
                try:
                    event[key] = next(pattern)
                except StopIteration:
                    return
            if (yield self.event_type(id_, **event)):
                return

    @property
    def is_infinite(self):
        for value in self._patterns.values():
            if isinstance(value, Pattern) and not value.is_infinite:
                return False
        return True


class UpdatePattern(Pattern):
    """
    Akin to SuperCollider's Pbindf.
    """

    def __init__(self, pattern, **patterns):
        self._pattern = pattern
        self._patterns = patterns

    def _iterate(self, state=None):
        event_iterator = iter(self._pattern)
        iterator_pairs = sorted(self._prepare_patterns().items())
        while True:
            try:
                event = next(event_iterator)
            except StopIteration:
                return
            template_dict = {}
            for key, key_iterator in iterator_pairs:
                try:
                    template_dict[key] = next(key_iterator)
                except StopIteration:
                    return
            event = new(event, **template_dict)
            if (yield event):
                return

    def _prepare_patterns(self):
        patterns = self._patterns.copy()
        for name, pattern in sorted(patterns.items()):
            if not isinstance(pattern, Pattern):
                pattern = SequencePattern([pattern], iterations=None)
            patterns[name] = iter(pattern)
        return patterns

    @property
    def is_infinite(self):
        for value in self._patterns.values():
            if isinstance(value, Pattern) and not value.is_infinite:
                return False
        return self._pattern.is_infinite


class ChainPattern(Pattern):
    """
    Akin to SuperCollider's Pchain.
    """

    def __init__(self, *patterns):
        self._patterns = tuple(patterns)

    def _iterate(self, state=None):
        patterns = [iter(_) for _ in self._patterns]
        while True:
            try:
                event = next(patterns[0])
            except StopIteration:
                return
            for pattern in patterns[1:]:
                try:
                    event = event.merge(next(pattern))
                except StopIteration:
                    return
            if (yield event):
                return

    @property
    def is_infinite(self):
        for value in self._patterns:
            if isinstance(value, Pattern) and not value.is_infinite:
                return False
        return True



import abc
import collections
import inspect
from collections.abc import Iterable, Sequence
from typing import Tuple

import uqbar.strings

import supriya
from supriya import BinaryOperator, SignalRange, UnaryOperator
from supriya.typing import UGenInputMap

from .mixins import UGenMethodMixin


class UGenMeta(abc.ABCMeta):

    initializer_template = uqbar.strings.normalize(
        """
    def __init__(
        self,
        {parameters_indent_one}
    ):
        {base_class_name}.__init__(
            self,
            {parameters_indent_two}
            )
    """
    )

    constructor_template = uqbar.strings.normalize(
        '''
    @classmethod
    def {rate_token}(
        cls,
        {parameters_indent_one}
    ):
        """
        Constructs a {rate_name_lower}-rate ``{ugen_name}`` unit generator graph.

        Returns unit generator graph.
        """
        calculation_rate = supriya.CalculationRate.{rate_name_upper}
        {validators}
        return cls._new_expanded(
            calculation_rate=calculation_rate,
            {parameters_indent_two}
        )
    '''
    )

    rateless_constructor_template = uqbar.strings.normalize(
        '''
    @classmethod
    def new(
        cls,
        {parameters_indent_one}
    ):
        """
        Constructs a ``{ugen_name}`` unit generator graph.

        Returns unit generator graph.
        """
        {validators}
        return cls._new_expanded(
            {parameters_indent_two}
        )
    '''
    )

    def __new__(metaclass, class_name, bases, namespace):
        ordered_input_names = namespace.get("_ordered_input_names")
        unexpanded_input_names = namespace.get("_unexpanded_input_names", ())
        valid_calculation_rates = namespace.get("_valid_calculation_rates", ())
        (
            default_channel_count,
            has_settable_channel_count,
        ) = UGenMeta.get_channel_count(namespace, bases)
        if isinstance(ordered_input_names, collections.OrderedDict):
            for name in ordered_input_names:
                if name in namespace:
                    continue
                namespace[name] = UGenMeta.make_property(
                    ugen_name=class_name,
                    input_name=name,
                    unexpanded=name in unexpanded_input_names,
                )
            if "__init__" not in namespace:
                function, string = UGenMeta.make_initializer(
                    ugen_name=class_name,
                    bases=bases,
                    parameters=ordered_input_names.copy(),
                    has_calculation_rate=bool(valid_calculation_rates),
                    default_channel_count=default_channel_count,
                    has_settable_channel_count=has_settable_channel_count,
                )
                namespace["__init__"] = function
                namespace["_init_source"] = string
            constructor_rates = {}
            if valid_calculation_rates:
                for rate in valid_calculation_rates:
                    if rate.token in namespace:
                        continue
                    constructor_rates[rate.token] = rate
            elif "new" not in namespace:
                constructor_rates["new"] = None
            for name, rate in constructor_rates.items():
                function, string = UGenMeta.make_constructor(
                    ugen_name=class_name,
                    bases=bases,
                    rate=rate,
                    parameters=ordered_input_names.copy(),
                    default_channel_count=default_channel_count,
                    has_settable_channel_count=has_settable_channel_count,
                )
                namespace[name] = function
                namespace["_{}_source".format(name)] = string
        return super().__new__(metaclass, class_name, bases, namespace)

    @staticmethod
    def get_channel_count(namespace, bases):
        default_channel_count = namespace.get("_default_channel_count")
        if default_channel_count is None:
            for base in bases:
                default_channel_count = getattr(base, "_default_channel_count")
                if default_channel_count is not None:
                    break
        has_settable_channel_count = namespace.get("_has_settable_channel_count")
        if has_settable_channel_count is None:
            for base in bases:
                has_settable_channel_count = getattr(
                    base, "_has_settable_channel_count"
                )
                if has_settable_channel_count is not None:
                    break
        return default_channel_count, has_settable_channel_count

    @staticmethod
    def compile(string, object_name, ugen_name, bases):
        namespace = {"supriya": supriya}
        namespace[bases[0].__name__] = bases[0]
        source_name = "<auto-generated> {}.py".format(ugen_name)
        try:
            code = compile(string, source_name, "exec")
            exec(code, namespace, namespace)
        except Exception:
            print(string)
            raise
        return namespace[object_name]

    @staticmethod
    def make_constructor(
        ugen_name,
        bases,
        rate,
        parameters,
        default_channel_count=False,
        has_settable_channel_count=False,
    ):
        validators = ""
        if default_channel_count and has_settable_channel_count:
            parameters["channel_count"] = int(default_channel_count)
            validators = ("\n" + (" " * 4)).join(
                [
                    "if channel_count < 1:",
                    "    raise ValueError('Channel count must be greater than zero')",
                ]
            )
        parameters_indent_one = (
            "\n    ".join(
                ["{}={},".format(key, value) for key, value in parameters.items()]
            )
            .replace("=inf,", "=float('inf'),")
            .replace("=-inf,", "=float('-inf'),")
        )
        parameters_indent_two = "\n        ".join(
            ["{}={},".format(key, key) for key, value in parameters.items()]
        )
        kwargs = dict(
            parameters_indent_one=parameters_indent_one,
            parameters_indent_two=parameters_indent_two,
            ugen_name=ugen_name,
            validators=validators,
        )
        if rate is None:
            string = UGenMeta.rateless_constructor_template.format(**kwargs)
        else:
            kwargs.update(
                rate_name_lower=rate.name.lower(),
                rate_name_upper=rate.name,
                rate_token=rate.token,
            )
            string = UGenMeta.constructor_template.format(**kwargs)
        object_name = "new" if rate is None else rate.token
        function = UGenMeta.compile(string, object_name, ugen_name, bases)
        return function, string

    @staticmethod
    def make_initializer(
        ugen_name,
        bases,
        parameters,
        has_calculation_rate=True,
        default_channel_count=False,
        has_settable_channel_count=False,
    ):
        if has_calculation_rate:
            new_parameters = collections.OrderedDict([("calculation_rate", None)])
            new_parameters.update(parameters)
            parameters = new_parameters
        if default_channel_count and has_settable_channel_count:
            parameters["channel_count"] = int(default_channel_count)
        parameters_indent_one = (
            "\n    ".join(
                ["{}={},".format(key, value) for key, value in parameters.items()]
            )
            .replace("=inf,", "=float('inf'),")
            .replace("=-inf,", "=float('-inf'),")
        )
        if has_settable_channel_count:
            parameters["channel_count"] = "channel_count"
        elif default_channel_count > 1:
            parameters["channel_count"] = int(default_channel_count)
        parameters_indent_two = "\n        ".join(
            [
                "{}={},".format(key, value if key == "channel_count" else key)
                for key, value in parameters.items()
            ]
        )
        string = UGenMeta.initializer_template.format(
            base_class_name=bases[0].__name__,
            parameters_indent_one=parameters_indent_one,
            parameters_indent_two=parameters_indent_two,
        )
        function = UGenMeta.compile(string, "__init__", ugen_name, bases)
        return function, string

    @staticmethod
    def make_property(ugen_name, input_name, unexpanded=False):
        doc = uqbar.strings.normalize(
            """
        Gets ``{input_name}`` of ``{ugen_name}``.

        Returns input.
        """
        ).format(ugen_name=ugen_name, input_name=input_name)
        if unexpanded:

            def getter(object_):
                index = tuple(object_._ordered_input_names).index(input_name)
                return tuple(object_._inputs[index:])

        else:

            def getter(object_):
                index = tuple(object_._ordered_input_names).index(input_name)
                return object_._inputs[index]

        return property(fget=getter, doc=doc)


class UGen(UGenMethodMixin, metaclass=UGenMeta):
    """
    A UGen.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_inputs", "_special_index", "_uuid")

    _default_channel_count = 1

    _has_settable_channel_count = False

    _has_done_flag = False

    _is_input = False

    _is_output = False

    _is_pure = False

    _is_width_first = False

    _ordered_input_names: UGenInputMap = None

    _signal_range: int = SignalRange.BIPOLAR

    _unexpanded_input_names: Tuple[str, ...] = ()

    _valid_calculation_rates: Tuple[int, ...] = ()

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, calculation_rate=None, special_index=0, **kwargs):
        import supriya.synthdefs

        calculation_rate = supriya.CalculationRate.from_expr(calculation_rate)
        if self._valid_calculation_rates:
            assert calculation_rate in self._valid_calculation_rates
        self._calculation_rate = calculation_rate
        self._inputs = []
        self._special_index = special_index
        ugenlike_prototype = (UGen, supriya.synthdefs.Parameter)
        server_id_prototype = (
            supriya.realtime.ServerObject,
            supriya.realtime.BusProxy,
            supriya.realtime.BufferProxy,
        )
        for input_name in self._ordered_input_names:
            input_value = None
            if input_name in kwargs:
                input_value = kwargs.pop(input_name)
            if isinstance(input_value, ugenlike_prototype):
                assert len(input_value) == 1
                input_value = input_value[0]
            elif isinstance(input_value, server_id_prototype):
                input_value = int(input_value)
            if self._is_unexpanded_input_name(input_name):
                if not isinstance(input_value, Sequence):
                    input_value = (input_value,)
                if isinstance(input_value, Sequence):
                    input_value = tuple(input_value)
                elif not self._is_valid_input(input_value):
                    raise ValueError(input_name, input_value)
            elif not self._is_valid_input(input_value):
                raise ValueError(input_name, input_value)
            self._configure_input(input_name, input_value)
        if kwargs:
            raise ValueError(kwargs)
        assert all(
            isinstance(_, (supriya.synthdefs.OutputProxy, float)) for _ in self.inputs
        )
        self._validate_inputs()
        self._uuid = None
        if supriya.synthdefs.SynthDefBuilder._active_builders:
            builder = supriya.synthdefs.SynthDefBuilder._active_builders[-1]
            self._uuid = builder._uuid
        self._check_inputs_share_same_uuid()
        if self._uuid is not None:
            builder._add_ugens(self)

    ### SPECIAL METHODS ###

    def __getitem__(self, i):
        """
        Gets output proxy at index `i`.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar()
            >>> ugen[0]
            SinOsc.ar()[0]

        Returns output proxy.
        """
        return self._get_output_proxy(i)

    def __len__(self):
        """
        Gets number of ugen outputs.

        Returns integer.
        """
        return getattr(self, "_channel_count", self._default_channel_count)

    def __repr__(self):
        """
        Gets interpreter representation of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar()
            >>> repr(ugen)
            'SinOsc.ar()'

        ::

            >>> ugen = supriya.ugens.WhiteNoise.kr()
            >>> repr(ugen)
            'WhiteNoise.kr()'

        ::

            >>> ugen = supriya.ugens.Rand.ir()
            >>> repr(ugen)
            'Rand.ir()'

        Returns string.
        """
        import supriya.synthdefs

        if self.calculation_rate == supriya.CalculationRate.DEMAND:
            return "{}()".format(type(self).__name__)
        calculation_abbreviations = {
            supriya.CalculationRate.AUDIO: "ar",
            supriya.CalculationRate.CONTROL: "kr",
            supriya.CalculationRate.SCALAR: "ir",
        }
        string = "{}.{}()".format(
            type(self).__name__, calculation_abbreviations[self.calculation_rate]
        )
        return string

    ### PRIVATE METHODS ###

    @staticmethod
    def _as_audio_rate_input(expr):
        import supriya.synthdefs
        import supriya.ugens

        if isinstance(expr, (int, float)):
            if expr == 0:
                return supriya.ugens.Silence.ar()
            return supriya.ugens.DC.ar(expr)
        elif isinstance(expr, (UGen, supriya.synthdefs.OutputProxy)):
            if expr.calculation_rate == supriya.CalculationRate.AUDIO:
                return expr
            return supriya.ugens.K2A.ar(source=expr)
        elif isinstance(expr, Iterable):
            return supriya.synthdefs.UGenArray(
                UGen._as_audio_rate_input(x) for x in expr
            )
        raise ValueError(expr)

    def _add_constant_input(self, value):
        self._inputs.append(float(value))

    def _add_ugen_input(self, ugen, output_index=None):
        import supriya.synthdefs

        # if isinstance(ugen, supriya.synthdefs.Parameter):
        #    output_proxy = ugen
        if isinstance(ugen, supriya.synthdefs.OutputProxy):
            output_proxy = ugen
        else:
            output_proxy = supriya.synthdefs.OutputProxy(
                output_index=output_index, source=ugen
            )
        self._inputs.append(output_proxy)

    def _check_inputs_share_same_uuid(self):
        import supriya.synthdefs

        for input_ in self.inputs:
            if not isinstance(input_, supriya.synthdefs.OutputProxy):
                continue
            if input_.source._uuid != self._uuid:
                message = "UGen input in different scope: {!r}"
                message = message.format(input_.source)
                raise ValueError(message)

    def _check_rate_same_as_first_input_rate(self):
        import supriya.synthdefs

        first_input_rate = supriya.CalculationRate.from_expr(self.inputs[0])
        return self.calculation_rate == first_input_rate

    def _check_range_of_inputs_at_audio_rate(self, start=None, stop=None):
        import supriya.synthdefs

        if self.calculation_rate != supriya.CalculationRate.AUDIO:
            return True
        for input_ in self.inputs[start:stop]:
            calculation_rate = supriya.CalculationRate.from_expr(input_)
            if calculation_rate != supriya.CalculationRate.AUDIO:
                return False
        return True

    def _configure_input(self, name, value):
        import supriya.synthdefs

        ugen_prototype = (
            supriya.synthdefs.OutputProxy,
            supriya.synthdefs.Parameter,
            UGen,
        )
        if hasattr(value, "__float__"):
            self._add_constant_input(float(value))
        elif isinstance(value, ugen_prototype):
            self._add_ugen_input(value._get_source(), value._get_output_number())
        elif isinstance(value, tuple):
            assert self._unexpanded_input_names
            assert name in self._unexpanded_input_names
            for x in value:
                if hasattr(x, "__float__"):
                    self._add_constant_input(float(x))
                elif isinstance(x, ugen_prototype):
                    self._add_ugen_input(x._get_source(), x._get_output_number())
                else:
                    raise Exception("{!r} {!r}".format(value, x))
        else:
            raise Exception(repr(value))

    @staticmethod
    def _expand_dictionary(dictionary, unexpanded_input_names=None):
        """
        Expands a dictionary into multichannel dictionaries.

        ::

            >>> dictionary = {"foo": 0, "bar": (1, 2), "baz": (3, 4, 5)}
            >>> result = supriya.synthdefs.UGen._expand_dictionary(dictionary)
            >>> for x in result:
            ...     sorted(x.items())
            ...
            [('bar', 1), ('baz', 3), ('foo', 0)]
            [('bar', 2), ('baz', 4), ('foo', 0)]
            [('bar', 1), ('baz', 5), ('foo', 0)]

        ::

            >>> dictionary = {"bus": (8, 9), "source": (1, 2, 3)}
            >>> result = supriya.synthdefs.UGen._expand_dictionary(
            ...     dictionary,
            ...     unexpanded_input_names=("source",),
            ... )
            >>> for x in result:
            ...     sorted(x.items())
            ...
            [('bus', 8), ('source', (1, 2, 3))]
            [('bus', 9), ('source', (1, 2, 3))]

        """
        import supriya.synthdefs

        dictionary = dictionary.copy()
        cached_unexpanded_inputs = {}
        if unexpanded_input_names is not None:
            for input_name in unexpanded_input_names:
                if input_name not in dictionary:
                    continue
                cached_unexpanded_inputs[input_name] = dictionary[input_name]
                del dictionary[input_name]
        maximum_length = 1
        result = []
        prototype = (Sequence, UGen, supriya.synthdefs.Parameter)
        for name, value in dictionary.items():
            if isinstance(value, prototype) and not isinstance(value, str):
                maximum_length = max(maximum_length, len(value))
        for i in range(maximum_length):
            result.append({})
            for name, value in dictionary.items():
                if isinstance(value, prototype) and not isinstance(value, str):
                    value = value[i % len(value)]
                    result[i][name] = value
                else:
                    result[i][name] = value
        for expanded_inputs in result:
            expanded_inputs.update(cached_unexpanded_inputs)
        return result

    def _get_done_action(self):
        import supriya.synthdefs

        if "done_action" not in self._ordered_input_names:
            return None
        return supriya.DoneAction.from_expr(int(self.done_action))

    @staticmethod
    def _get_method_for_rate(cls, calculation_rate):
        import supriya.synthdefs

        calculation_rate = supriya.CalculationRate.from_expr(calculation_rate)
        if calculation_rate == supriya.CalculationRate.AUDIO:
            return cls.ar
        elif calculation_rate == supriya.CalculationRate.CONTROL:
            return cls.kr
        elif calculation_rate == supriya.CalculationRate.SCALAR:
            if hasattr(cls, "ir"):
                return cls.ir
            return cls.kr
        return cls.new

    def _get_output_number(self):
        return 0

    def _get_outputs(self):
        return [self.calculation_rate] * len(self)

    def _get_source(self):
        return self

    def _is_unexpanded_input_name(self, input_name):
        if self._unexpanded_input_names:
            if input_name in self._unexpanded_input_names:
                return True
        return False

    def _is_valid_input(self, input_value):
        import supriya.synthdefs

        if isinstance(input_value, supriya.synthdefs.OutputProxy):
            return True
        elif hasattr(input_value, "__float__"):
            return True
        return False

    @classmethod
    def _new_expanded(cls, special_index=0, **kwargs):
        import supriya.synthdefs

        get_signature = inspect.signature
        signature = get_signature(cls.__init__)
        input_dicts = UGen._expand_dictionary(
            kwargs, unexpanded_input_names=cls._unexpanded_input_names
        )
        ugens = []
        has_custom_special_index = "special_index" in signature.parameters
        for input_dict in input_dicts:
            if has_custom_special_index:
                ugen = cls._new_single(special_index=special_index, **input_dict)
            else:
                ugen = cls._new_single(**input_dict)
            ugens.append(ugen)
        if len(ugens) == 1:
            return ugens[0]
        return supriya.synthdefs.UGenArray(ugens)

    @classmethod
    def _new_single(cls, **kwargs):
        ugen = cls(**kwargs)
        return ugen

    def _optimize_graph(self, sort_bundles):
        if self._is_pure:
            self._perform_dead_code_elimination(sort_bundles)

    def _perform_dead_code_elimination(self, sort_bundles):
        sort_bundle = sort_bundles.get(self, None)
        if not sort_bundle or sort_bundle.descendants:
            return
        del sort_bundles[self]
        for antecedent in tuple(sort_bundle.antecedents):
            antecedent_bundle = sort_bundles.get(antecedent, None)
            if not antecedent_bundle:
                continue
            antecedent_bundle.descendants.remove(self)
            antecedent._optimize_graph(sort_bundles)

    def _validate_inputs(self):
        pass

    ### PRIVATE PROPERTIES ###

    @property
    def _has_done_action(self):
        return "done_action" in self._ordered_input_names

    ### PUBLIC PROPERTIES ###

    @property
    def calculation_rate(self):
        """
        Gets calculation-rate of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar(
            ...     frequency=supriya.ugens.WhiteNoise.kr(),
            ...     phase=0.5,
            ... )
            >>> ugen.calculation_rate
            CalculationRate.AUDIO

        Returns calculation-rate.
        """
        return self._calculation_rate

    @property
    def has_done_flag(self):
        return self._has_done_flag

    @property
    def inputs(self):
        """
        Gets inputs of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar(
            ...     frequency=supriya.ugens.WhiteNoise.kr(),
            ...     phase=0.5,
            ... )
            >>> for input_ in ugen.inputs:
            ...     input_
            ...
            WhiteNoise.kr()[0]
            0.5

        Returns tuple.
        """
        return tuple(self._inputs)

    @property
    def is_input_ugen(self):
        return self._is_input

    @property
    def is_output_ugen(self):
        return self._is_output

    @property
    def outputs(self):
        """
        Gets outputs of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar(
            ...     frequency=supriya.ugens.WhiteNoise.kr(),
            ...     phase=0.5,
            ... )
            >>> ugen.outputs
            (CalculationRate.AUDIO,)

        Returns tuple.
        """
        return tuple(self._get_outputs())

    @property
    def signal_range(self):
        """
        Gets signal range of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar()
            >>> ugen.signal_range
            SignalRange.BIPOLAR

        A bipolar signal range indicates that the ugen generates signals above
        and below zero.

        A unipolar signal range indicates that the ugen only generates signals
        of 0 or greater.

        Returns signal range.
        """
        return self._signal_range

    @property
    def special_index(self):
        """
        Gets special index of ugen.

        ::

            >>> ugen = supriya.ugens.SinOsc.ar(
            ...     frequency=supriya.ugens.WhiteNoise.kr(),
            ...     phase=0.5,
            ... )
            >>> ugen.special_index
            0

        The `special index` of most ugens will be 0. SuperColliders's synth
        definition file format uses the special index to store the operator id
        for binary and unary operator ugens, and the parameter index of
        controls.

        Returns integer.
        """
        return self._special_index


class MultiOutUGen(UGen):
    """
    Abstract base class for ugens with multiple outputs.
    """

    ### INTIALIZER ###

    @abc.abstractmethod
    def __init__(
        self, calculation_rate=None, special_index=0, channel_count=1, **kwargs
    ):
        self._channel_count = int(channel_count)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            special_index=special_index,
            **kwargs,
        )

    ### SPECIAL METHODS ###

    def __len__(self):
        """
        Gets number of ugen outputs.

        Returns integer.
        """
        return self._channel_count

    ### PRIVATE METHODS ###

    @classmethod
    def _new_expanded(cls, special_index=0, **kwargs):
        import supriya.synthdefs
        import supriya.ugens

        ugen = super(MultiOutUGen, cls)._new_expanded(
            special_index=special_index, **kwargs
        )
        output_proxies = []
        if isinstance(ugen, UGen):
            output_proxies.extend(ugen[:])
        else:
            for x in ugen:
                output_proxies.extend(x[:])
        if len(output_proxies) == 1:
            return output_proxies[0]
        result = supriya.synthdefs.UGenArray(output_proxies)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def channel_count(self):
        """
        Gets channel count of multi-output ugen.

        Returns integer.
        """
        return self._channel_count


class PureUGen(UGen):
    """
    Abstract base class for ugens with no side-effects.

    These ugens may be optimized out of ugen graphs during SynthDef
    compilation.
    """

    ### CLASS VARIABLES ###

    _is_pure = True

    ### PRIVATE METHODS ###

    def _optimize_graph(self, sort_bundles):
        self._perform_dead_code_elimination(sort_bundles)


class PureMultiOutUGen(MultiOutUGen):
    """
    Abstract base class for multi-output ugens with no side-effects.

    These ugens may be optimized out of ugen graphs during SynthDef
    compilation.
    """

    def _optimize_graph(self, sort_bundles):
        self._perform_dead_code_elimination(sort_bundles)


class WidthFirstUGen(UGen):
    """
    Abstract base class for UGens with a width-first sort order.
    """

    ### CLASS VARIABLES ###

    _is_width_first = True

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, calculation_rate=None, special_index=0, **kwargs):
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            special_index=special_index,
            **kwargs,
        )


class UnaryOpUGen(PureUGen):
    """
    A unary operator ugen, created by applying a unary operator to a ugen.

    ::

        >>> ugen = supriya.ugens.SinOsc.ar()
        >>> unary_op_ugen = abs(ugen)
        >>> unary_op_ugen
        UnaryOpUGen.ar()

    ::

        >>> unary_op_ugen.operator
        UnaryOperator.ABSOLUTE_VALUE

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("source", None)])

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, source=None, special_index=None):
        PureUGen.__init__(
            self,
            calculation_rate=calculation_rate,
            source=source,
            special_index=special_index,
        )

    ### PUBLIC PROPERTIES ###

    @property
    def operator(self):
        """
        Gets operator of UnaryOpUgen.

        ::

            >>> source = supriya.ugens.SinOsc.ar()
            >>> unary_op_ugen = -source
            >>> unary_op_ugen.operator
            UnaryOperator.NEGATIVE

        Returns unary operator.
        """
        return UnaryOperator(self.special_index)


class BinaryOpUGen(PureUGen):
    """
    A binary operator ugen, created by applying a binary operator to two
    ugens.

    ::

        >>> left_operand = supriya.ugens.SinOsc.ar()
        >>> right_operand = supriya.ugens.WhiteNoise.kr()
        >>> binary_op_ugen = left_operand * right_operand
        >>> binary_op_ugen
        BinaryOpUGen.ar()

    ::

        >>> binary_op_ugen.operator
        BinaryOperator.MULTIPLICATION

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("left", None), ("right", None)])

    ### INITIALIZER ###

    def __init__(
        self, calculation_rate=None, special_index=None, left=None, right=None
    ):
        PureUGen.__init__(
            self,
            calculation_rate=calculation_rate,
            special_index=special_index,
            left=left,
            right=right,
        )

    ### PRIVATE METHODS ###

    @classmethod
    def _new_single(
        cls, calculation_rate=None, special_index=None, left=None, right=None
    ):
        a = left
        b = right
        if special_index == BinaryOperator.MULTIPLICATION:
            if a == 0:
                return 0
            if b == 0:
                return 0
            if a == 1:
                return b
            if a == 1:
                return -b
            if b == 1:
                return a
            if b == -1:
                return -a
        if special_index == BinaryOperator.ADDITION:
            if a == 0:
                return b
            if b == 0:
                return a
        if special_index == BinaryOperator.SUBTRACTION:
            if a == 0:
                return -b
            if b == 0:
                return a
        if special_index == BinaryOperator.FLOAT_DIVISION:
            if b == 1:
                return a
            if b == -1:
                return -a
        ugen = cls(
            calculation_rate=calculation_rate,
            special_index=special_index,
            left=a,
            right=b,
        )
        return ugen

    ### PUBLIC PROPERTIES ###

    @property
    def operator(self):
        """
        Gets operator of BinaryOpUgen.

        ::

            >>> left = supriya.ugens.SinOsc.ar()
            >>> right = supriya.ugens.WhiteNoise.kr()
            >>> binary_op_ugen = left / right
            >>> binary_op_ugen.operator
            BinaryOperator.FLOAT_DIVISION

        Returns binary operator.
        """
        return BinaryOperator(self.special_index)


class PseudoUGen:
    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError



import supriya


def _build_synthdef():
    with supriya.SynthDefBuilder(amplitude=0, bus=0, frequency=440) as builder:
        supriya.ugens.Out.ar(
            bus=builder["bus"],
            source=supriya.ugens.SinOsc.ar(frequency=builder["frequency"])
            * builder["amplitude"],
        )
    return builder.build(name="simple_sine")


simple_sine = _build_synthdef()


__all__ = ["simple_sine"]



import typing
from collections.abc import Sequence

import supriya.osc
from supriya.enums import RequestId

from .bases import Request


class MoveRequest(Request):

    ### CLASS VARIABLES ###

    class NodeIdPair(typing.NamedTuple):
        node_id: int
        target_node_id: int

    ### INITIALIZER ###

    def __init__(self, node_id_pairs=None):
        Request.__init__(self)
        if node_id_pairs:
            if not isinstance(node_id_pairs, Sequence):
                node_id_pairs = [node_id_pairs]
            node_id_pairs = list(node_id_pairs)
            for i, x in enumerate(node_id_pairs):
                if not isinstance(x, self.NodeIdPair):
                    node_id_pairs[i] = self.NodeIdPair(*x)
            node_id_pairs = tuple(node_id_pairs)
        self._node_id_pairs = node_id_pairs

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.node_id_pairs:
            for node_id_pair in self.node_id_pairs:
                target_node_id = self._sanitize_node_id(
                    node_id_pair.target_node_id, with_placeholders
                )
                node_id = self._sanitize_node_id(
                    node_id_pair.node_id, with_placeholders
                )
                if self._target_first:
                    contents.append(target_node_id)
                    contents.append(node_id)
                else:
                    contents.append(node_id)
                    contents.append(target_node_id)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id_pairs(self):
        return self._node_id_pairs


class GroupHeadRequest(MoveRequest):
    """
    A /g_head request.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)
        >>> synth = supriya.Synth().allocate(server)
        >>> group.extend([supriya.Group(), supriya.Group()])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1000 group
                    1002 group
                    1003 group

    ::

        >>> request = supriya.commands.GroupHeadRequest([(synth, group)])
        >>> request.to_osc()
        OscMessage('/g_head', 1000, 1001)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     response = request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> response

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/g_head', 1000, 1001))
        ('S', OscMessage('/sync', 1))
        ('R', OscMessage('/n_move', 1001, 1000, -1, 1002, 0))
        ('R', OscMessage('/synced', 1))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1002 group
                    1003 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1002 group
                    1003 group

    """

    ### CLASS VARIABLES ###

    _target_first = True

    request_id = RequestId.GROUP_HEAD


class GroupTailRequest(MoveRequest):
    """
    A /g_tail request.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)
        >>> synth = supriya.Synth().allocate(server)
        >>> group.extend([supriya.Group(), supriya.Group()])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1000 group
                    1002 group
                    1003 group

    ::

        >>> request = supriya.commands.GroupTailRequest([(synth, group)])
        >>> request.to_osc()
        OscMessage('/g_tail', 1000, 1001)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     response = request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> response

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/g_tail', 1000, 1001))
        ('S', OscMessage('/sync', 1))
        ('R', OscMessage('/n_move', 1001, 1000, 1003, -1, 0))
        ('R', OscMessage('/synced', 1))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1002 group
                    1003 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                    1002 group
                    1003 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

    """

    ### CLASS VARIABLES ###

    _target_first = True

    request_id = RequestId.GROUP_TAIL


class NodeAfterRequest(MoveRequest):
    """
    An /n_after request.

    ::

        >>> server = supriya.Server().boot()
        >>> group_a = supriya.Group([supriya.Group(), supriya.Group()])
        >>> group_b = supriya.Group([supriya.Group(), supriya.Group()])
        >>> synth_a = supriya.Synth()
        >>> synth_b = supriya.Synth()
        >>> server.default_group.extend([synth_a, synth_b, group_a, group_b])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1002 group
                    1003 group
                    1004 group
                1005 group
                    1006 group
                    1007 group

    ::

        >>> request = supriya.commands.NodeAfterRequest(
        ...     [
        ...         [synth_a, group_a[-1]],
        ...         [synth_b, group_b],
        ...     ]
        ... )
        >>> request.to_osc()
        OscMessage('/n_after', 1000, 1004, 1001, 1005)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/n_after', 1000, 1004, 1001, 1005))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/n_move', 1000, 1002, 1004, -1, 0))
        ('R', OscMessage('/n_move', 1001, 1, 1005, -1, 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1002 group
                    1003 group
                    1004 group
                    1000 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1005 group
                    1006 group
                    1007 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1002 group
                    1003 group
                    1004 group
                    1000 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1005 group
                    1006 group
                    1007 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

    """

    ### CLASS VARIABLES ###

    _target_first = False

    request_id = RequestId.NODE_AFTER


class NodeBeforeRequest(MoveRequest):
    """
    An /n_before request.

    ::

        >>> server = supriya.Server().boot()
        >>> group_a = supriya.Group([supriya.Group(), supriya.Group()])
        >>> group_b = supriya.Group([supriya.Group(), supriya.Group()])
        >>> synth_a = supriya.Synth()
        >>> synth_b = supriya.Synth()
        >>> server.default_group.extend([synth_a, synth_b, group_a, group_b])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1002 group
                    1003 group
                    1004 group
                1005 group
                    1006 group
                    1007 group

    ::

        >>> request = supriya.commands.NodeBeforeRequest(
        ...     [
        ...         [synth_a, group_a[0]],
        ...         [synth_b, group_b],
        ...     ]
        ... )
        >>> request.to_osc()
        OscMessage('/n_before', 1000, 1003, 1001, 1005)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/n_before', 1000, 1003, 1001, 1005))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/n_move', 1000, 1002, -1, 1003, 0))
        ('R', OscMessage('/n_move', 1001, 1, 1002, 1005, 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1002 group
                    1000 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1003 group
                    1004 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1005 group
                    1006 group
                    1007 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1002 group
                    1000 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1003 group
                    1004 group
                1001 default
                    out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                1005 group
                    1006 group
                    1007 group

    """

    ### CLASS VARIABLES ###

    _target_first = False

    request_id = RequestId.NODE_BEFORE



from .clap import clap  # noqa
from .default import default  # noqa
from .kick import kick  # noqa
from .multiband_compressor import multiband_compressor  # noqa
from .simple_sine import simple_sine  # noqa
from .sweep_filter import sweep_filter  # noqa
from .system_synthdefs import *  # noqa
from .test import test  # noqa



import collections
import copy
import hashlib
import os
import pathlib
import shutil
import subprocess
import tempfile

import yaml

from supriya import (
    BinaryOperator,
    CalculationRate,
    ParameterRate,
    UnaryOperator,
    sclang,
)
from supriya.system import SupriyaObject

from .bases import BinaryOpUGen, UGen, UnaryOpUGen, WidthFirstUGen
from .compilers import SynthDefCompiler
from .controls import AudioControl, Control, LagControl, Parameter, TrigControl
from .grapher import SynthDefGrapher
from .mixins import OutputProxy, UGenMethodMixin


class SynthDef:
    """
    A synth definition.

    ::

        >>> import supriya.synthdefs
        >>> import supriya.ugens
        >>> with supriya.synthdefs.SynthDefBuilder(frequency=440) as builder:
        ...     sin_osc = supriya.ugens.SinOsc.ar(frequency=builder["frequency"])
        ...     out = supriya.ugens.Out.ar(bus=0, source=sin_osc)
        ...
        >>> synthdef = builder.build()

    ::

        >>> supriya.graph(synthdef)  # doctest: +SKIP

    ::

        >>> import supriya.realtime
        >>> server = supriya.Server().boot()

    ::

        >>> synthdef.allocate(server=server)
        <SynthDef: 9c4eb4778dc0faf39459fa8a5cd45c19>

    ::

        >>> synthdef in server
        True

    ::

        >>> synthdef.free(server)

    ::

        >>> synthdef in server
        False

    ::

        >>> server.quit()
        <Server: offline>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_compiled_ugen_graph",
        "_constants",
        "_control_ugens",
        "_indexed_parameters",
        "_name",
        "_ugens",
    )

    ### INITIALIZER ###

    def __init__(self, ugens, name=None, optimize=True, parameter_names=None, **kwargs):
        self._name = name
        ugens = list(copy.deepcopy(ugens))
        assert all(isinstance(_, UGen) for _ in ugens)
        ugens = self._cleanup_pv_chains(ugens)
        ugens = self._cleanup_local_bufs(ugens)
        if optimize:
            ugens = self._optimize_ugen_graph(ugens)
        ugens = self._sort_ugens_topologically(ugens)
        self._ugens = tuple(ugens)
        self._constants = self._collect_constants(self._ugens)
        self._control_ugens = self._collect_control_ugens(self._ugens)
        self._indexed_parameters = self._collect_indexed_parameters(
            self._control_ugens, parameter_names=parameter_names
        )
        self._compiled_ugen_graph = SynthDefCompiler.compile_ugen_graph(self)

    ### SPECIAL METHODS ###

    def __eq__(self, expr):
        if type(expr) != type(self):
            return False
        if expr.name != self.name:
            return False
        if expr._compiled_ugen_graph != self._compiled_ugen_graph:
            return False
        return True

    def __graph__(self):
        r"""
        Graphs SynthDef.

        ::

            >>> with supriya.synthdefs.SynthDefBuilder(frequency=440) as builder:
            ...     sin_osc = supriya.ugens.SinOsc.ar(frequency=builder["frequency"])
            ...     out = supriya.ugens.Out.ar(bus=0, source=sin_osc)
            ...
            >>> synthdef = builder.build()
            >>> print(format(synthdef.__graph__(), "graphviz"))
            digraph synthdef_... {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    ranksep=1,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    shape=Mrecord,
                    style="filled, rounded"];
                edge [penwidth=2];
                ugen_0 [fillcolor=lightgoldenrod2,
                    label="<f_0> Control\n(control) | { { <f_1_0_0> frequency:\n440.0 } }"];
                ugen_1 [fillcolor=lightsteelblue2,
                    label="<f_0> SinOsc\n(audio) | { { <f_1_0_0> frequency | <f_1_0_1> phase:\n0.0 } | { <f_1_1_0> 0 } }"];
                ugen_2 [fillcolor=lightsteelblue2,
                    label="<f_0> Out\n(audio) | { { <f_1_0_0> bus:\n0.0 | <f_1_0_1> source } }"];
                ugen_0:f_1_0_0:e -> ugen_1:f_1_0_0:w [color=goldenrod];
                ugen_1:f_1_1_0:e -> ugen_2:f_1_0_1:w [color=steelblue];
            }

        Returns Graphviz graph.
        """
        return SynthDefGrapher.graph(self)

    def __hash__(self):
        hash_values = (type(self), self._name, self._compiled_ugen_graph)
        return hash(hash_values)

    def __repr__(self):
        return "<{}: {}>".format(type(self).__name__, self.actual_name)

    def __str__(self):
        """
        Gets string representation of synth definition.

        ::

            >>> import supriya.synthdefs
            >>> import supriya.ugens

        ::

            >>> with supriya.synthdefs.SynthDefBuilder() as builder:
            ...     sin_one = supriya.ugens.SinOsc.ar()
            ...     sin_two = supriya.ugens.SinOsc.ar(frequency=443)
            ...     source = sin_one + sin_two
            ...     out = supriya.ugens.Out.ar(bus=0, source=source)
            ...
            >>> synthdef = builder.build(name="test")

        ::

            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: test
                ugens:
                -   SinOsc.ar/0:
                        frequency: 440.0
                        phase: 0.0
                -   SinOsc.ar/1:
                        frequency: 443.0
                        phase: 0.0
                -   BinaryOpUGen(ADDITION).ar:
                        left: SinOsc.ar/0[0]
                        right: SinOsc.ar/1[0]
                -   Out.ar:
                        bus: 0.0
                        source[0]: BinaryOpUGen(ADDITION).ar[0]

        Returns string.
        """

        def get_ugen_names():
            grouped_ugens = {}
            named_ugens = {}
            for ugen in self._ugens:
                key = (type(ugen), ugen.calculation_rate, ugen.special_index)
                grouped_ugens.setdefault(key, []).append(ugen)
            for ugen in self._ugens:
                parts = [type(ugen).__name__]
                if isinstance(ugen, BinaryOpUGen):
                    ugen_op = BinaryOperator.from_expr(ugen.special_index)
                    parts.append("(" + ugen_op.name + ")")
                elif isinstance(ugen, UnaryOpUGen):
                    ugen_op = UnaryOperator.from_expr(ugen.special_index)
                    parts.append("(" + ugen_op.name + ")")
                parts.append("." + ugen.calculation_rate.token)
                key = (type(ugen), ugen.calculation_rate, ugen.special_index)
                related_ugens = grouped_ugens[key]
                if len(related_ugens) > 1:
                    parts.append("/{}".format(related_ugens.index(ugen)))
                named_ugens[ugen] = "".join(parts)
            return named_ugens

        def get_parameter_name(input_, output_index=0):
            if isinstance(input_, Parameter):
                return ":{}".format(input_.name)
            elif isinstance(input_, Control):
                # Handle array-like parameters
                value_index = 0
                for parameter in input_.parameters:
                    values = parameter.value
                    if isinstance(values, float):
                        values = [values]
                    for i in range(len(values)):
                        if value_index != output_index:
                            value_index += 1
                            continue
                        elif len(values) == 1:
                            return ":{}".format(parameter.name)
                        else:
                            return ":{}[{}]".format(parameter.name, i)
            return ""

        ugens = []
        named_ugens = get_ugen_names()
        for ugen in self._ugens:
            ugen_dict = {}
            ugen_name = named_ugens[ugen]
            for i, input_ in enumerate(ugen.inputs):
                if i < len(ugen._ordered_input_names):
                    argument_name = tuple(ugen._ordered_input_names)[i]
                else:
                    argument_name = tuple(ugen._ordered_input_names)[-1]
                if (
                    ugen._unexpanded_input_names
                    and argument_name in ugen._unexpanded_input_names
                ):
                    unexpanded_index = i - tuple(ugen._ordered_input_names).index(
                        argument_name
                    )
                    argument_name += "[{}]".format(unexpanded_index)
                if isinstance(input_, float):
                    value = input_
                else:
                    output_index = 0
                    if isinstance(input_, OutputProxy):
                        output_index = input_.output_index
                        input_ = input_.source
                    input_name = named_ugens[input_]
                    value = "{}[{}{}]".format(
                        input_name,
                        output_index,
                        get_parameter_name(input_, output_index),
                    )
                ugen_dict[argument_name] = value
            if not ugen_dict:
                ugen_dict = None
            ugens.append({ugen_name: ugen_dict})

        result = {
            "synthdef": {
                "name": self.actual_name,
                # 'hash': self.anonymous_name,
                "ugens": ugens,
            }
        }
        return yaml.dump(result, default_flow_style=False, indent=4).rstrip()

    ### PRIVATE METHODS ###

    @staticmethod
    def _allocate_synthdefs(synthdefs, server):
        # TODO: Should sync be configurable here?
        import supriya.commands

        d_recv_synthdef_groups = []
        d_recv_synth_group = []
        current_total = 0
        d_load_synthdefs = []
        if not synthdefs:
            return
        for synthdef in synthdefs:
            # synthdef._register_with_local_server(server=server)
            compiled = synthdef.compile()
            if 8192 < len(compiled):
                d_load_synthdefs.append(synthdef)
            elif current_total + len(compiled) < 8192:
                d_recv_synth_group.append(synthdef)
                current_total += len(compiled)
            else:
                d_recv_synthdef_groups.append(d_recv_synth_group)
                d_recv_synth_group = [synthdef]
                current_total = len(compiled)
        if d_recv_synth_group:
            d_recv_synthdef_groups.append(d_recv_synth_group)
        for d_recv_synth_group in d_recv_synthdef_groups:
            d_recv_request = supriya.commands.SynthDefReceiveRequest(
                synthdefs=tuple(d_recv_synth_group)
            )

            d_recv_request.communicate(server=server, sync=True)
        if d_load_synthdefs:
            temp_directory_path = tempfile.mkdtemp()
            for synthdef in d_load_synthdefs:
                file_name = "{}.scsyndef".format(synthdef.actual_name)
                file_path = os.path.join(temp_directory_path, file_name)
                with open(file_path, "wb") as file_pointer:
                    file_pointer.write(synthdef.compile())
            d_load_dir_request = supriya.commands.SynthDefLoadDirectoryRequest(
                directory_path=temp_directory_path
            )
            d_load_dir_request.communicate(server=server, sync=True)
            shutil.rmtree(temp_directory_path)

    @staticmethod
    def _build_control_mapping(parameters):
        control_mapping = collections.OrderedDict()
        scalar_parameters = []
        trigger_parameters = []
        audio_parameters = []
        control_parameters = []
        mapping = {
            ParameterRate.AUDIO: audio_parameters,
            ParameterRate.CONTROL: control_parameters,
            ParameterRate.SCALAR: scalar_parameters,
            ParameterRate.TRIGGER: trigger_parameters,
        }
        for parameter in parameters:
            mapping[parameter.parameter_rate].append(parameter)
        for filtered_parameters in mapping.values():
            filtered_parameters.sort(key=lambda x: x.name)
        control_ugens = []
        indexed_parameters = []
        starting_control_index = 0
        if scalar_parameters:
            control = Control(
                parameters=scalar_parameters,
                calculation_rate=CalculationRate.SCALAR,
                starting_control_index=starting_control_index,
            )
            control_ugens.append(control)
            for parameter in scalar_parameters:
                indexed_parameters.append((starting_control_index, parameter))
                starting_control_index += len(parameter)
            for i, output_proxy in enumerate(control._get_parameter_output_proxies()):
                control_mapping[output_proxy] = control[i]
        if trigger_parameters:
            control = TrigControl(
                parameters=trigger_parameters,
                starting_control_index=starting_control_index,
            )
            control_ugens.append(control)
            for parameter in trigger_parameters:
                indexed_parameters.append((starting_control_index, parameter))
                starting_control_index += len(parameter)
            for i, output_proxy in enumerate(control._get_parameter_output_proxies()):
                control_mapping[output_proxy] = control[i]
        if audio_parameters:
            control = AudioControl(
                parameters=audio_parameters,
                starting_control_index=starting_control_index,
            )
            control_ugens.append(control)
            for parameter in audio_parameters:
                indexed_parameters.append((starting_control_index, parameter))
                starting_control_index += len(parameter)
            for i, output_proxy in enumerate(control._get_parameter_output_proxies()):
                control_mapping[output_proxy] = control[i]
        if control_parameters:
            if any(_.lag for _ in control_parameters):
                control = LagControl(
                    parameters=control_parameters,
                    calculation_rate=CalculationRate.CONTROL,
                    starting_control_index=starting_control_index,
                )
            else:
                control = Control(
                    parameters=control_parameters,
                    calculation_rate=CalculationRate.CONTROL,
                    starting_control_index=starting_control_index,
                )
            control_ugens.append(control)
            for parameter in control_parameters:
                indexed_parameters.append((starting_control_index, parameter))
                starting_control_index += len(parameter)
            for i, output_proxy in enumerate(control._get_parameter_output_proxies()):
                control_mapping[output_proxy] = control[i]
        control_ugens = tuple(control_ugens)
        indexed_parameters.sort(key=lambda pair: parameters.index(pair[1]))
        indexed_parameters = tuple(indexed_parameters)
        return control_ugens, control_mapping, indexed_parameters

    @staticmethod
    def _build_input_mapping(ugens):
        import supriya.ugens

        input_mapping = {}
        for ugen in ugens:
            if not isinstance(ugen, supriya.ugens.PV_ChainUGen):
                continue
            if isinstance(ugen, supriya.ugens.PV_Copy):
                continue
            for i, input_ in enumerate(ugen.inputs):
                if not isinstance(input_, OutputProxy):
                    continue
                source = input_.source
                if not isinstance(source, supriya.ugens.PV_ChainUGen):
                    continue
                if source not in input_mapping:
                    input_mapping[source] = []
                input_mapping[source].append((ugen, i))
        return input_mapping

    @staticmethod
    def _cleanup_local_bufs(ugens):
        import supriya.ugens

        local_bufs = []
        processed_ugens = []
        for ugen in ugens:
            if isinstance(ugen, supriya.ugens.MaxLocalBufs):
                continue
            if isinstance(ugen, supriya.ugens.LocalBuf):
                local_bufs.append(ugen)
            processed_ugens.append(ugen)
        if local_bufs:
            max_local_bufs = supriya.ugens.MaxLocalBufs(len(local_bufs))
            for local_buf in local_bufs:
                inputs = list(local_buf.inputs[:2])
                inputs.append(max_local_bufs[0])
                local_buf._inputs = tuple(inputs)
            index = processed_ugens.index(local_bufs[0])
            processed_ugens[index:index] = [max_local_bufs]
        return tuple(processed_ugens)

    @staticmethod
    def _cleanup_pv_chains(ugens):
        import supriya.ugens

        input_mapping = SynthDef._build_input_mapping(ugens)
        for antecedent, descendants in input_mapping.items():
            if len(descendants) == 1:
                continue
            for descendant, input_index in descendants[:-1]:
                fft_size = antecedent.fft_size
                new_buffer = supriya.ugens.LocalBuf(fft_size)
                pv_copy = supriya.ugens.PV_Copy(antecedent, new_buffer)
                inputs = list(descendant._inputs)
                inputs[input_index] = pv_copy[0]
                descendant._inputs = tuple(inputs)
                index = ugens.index(descendant)
                replacement = []
                if isinstance(fft_size, UGenMethodMixin):
                    replacement.append(fft_size)
                replacement.extend([new_buffer, pv_copy])
                ugens[index:index] = replacement
        return ugens

    @staticmethod
    def _collect_constants(ugens):
        constants = []
        for ugen in ugens:
            for input_ in ugen._inputs:
                if not isinstance(input_, float):
                    continue
                if input_ not in constants:
                    constants.append(input_)
        return tuple(constants)

    @staticmethod
    def _collect_control_ugens(ugens):
        control_ugens = tuple(_ for _ in ugens if isinstance(_, Control))
        return control_ugens

    @staticmethod
    def _collect_indexed_parameters(control_ugens, parameter_names=None):
        indexed_parameters = []
        parameters = {}
        for control_ugen in control_ugens:
            index = control_ugen.starting_control_index
            for parameter in control_ugen.parameters:
                parameters[parameter.name] = (index, parameter)
                index += len(parameter)
        parameter_names = parameter_names or sorted(parameters)
        for parameter_name in parameter_names:
            indexed_parameters.append(parameters[parameter_name])
        indexed_parameters = tuple(indexed_parameters)
        return indexed_parameters

    @staticmethod
    def _extract_parameters(ugens):
        parameters = set()
        for ugen in ugens:
            if isinstance(ugen, Parameter):
                parameters.add(ugen)
        ugens = tuple(ugen for ugen in ugens if ugen not in parameters)
        parameters = tuple(sorted(parameters, key=lambda x: x.name))
        return ugens, parameters

    @staticmethod
    def _initialize_topological_sort(ugens):
        ugens = list(ugens)
        sort_bundles = collections.OrderedDict()
        width_first_antecedents = []
        for ugen in ugens:
            sort_bundles[ugen] = UGenSortBundle(ugen, width_first_antecedents)
            if isinstance(ugen, WidthFirstUGen):
                width_first_antecedents.append(ugen)
        for ugen in ugens:
            sort_bundle = sort_bundles[ugen]
            sort_bundle._initialize_topological_sort(sort_bundles)
            sort_bundle.descendants[:] = sorted(
                sort_bundles[ugen].descendants, key=lambda x: ugens.index(ugen)
            )
        return sort_bundles

    @staticmethod
    def _optimize_ugen_graph(ugens):
        sort_bundles = SynthDef._initialize_topological_sort(ugens)
        for ugen in ugens:
            ugen._optimize_graph(sort_bundles)
        return tuple(sort_bundles)

    def _register_with_local_server(self, server):
        synthdef_name = self.actual_name
        server._synthdefs[synthdef_name] = self

    @staticmethod
    def _remap_controls(ugens, control_mapping):
        for ugen in ugens:
            inputs = list(ugen.inputs)
            for i, input_ in enumerate(inputs):
                if input_ in control_mapping:
                    output_proxy = control_mapping[input_]
                    inputs[i] = output_proxy
            ugen._inputs = tuple(inputs)

    @staticmethod
    def _sort_ugens_topologically(ugens):
        sort_bundles = SynthDef._initialize_topological_sort(ugens)
        available_ugens = []
        for ugen in reversed(ugens):
            sort_bundles[ugen]._make_available(available_ugens)
        out_stack = []
        while available_ugens:
            available_ugen = available_ugens.pop()
            sort_bundles[available_ugen]._schedule(
                available_ugens, out_stack, sort_bundles
            )
        return out_stack

    ### PUBLIC METHODS ###

    def allocate(self, server):
        self._allocate_synthdefs((self,), server)
        return self

    def compile(self, use_anonymous_name=False):
        from supriya.synthdefs import SynthDefCompiler

        synthdefs = [self]
        result = SynthDefCompiler.compile_synthdefs(
            synthdefs, use_anonymous_names=use_anonymous_name
        )
        return result

    def free(self, server):
        import supriya.commands

        assert self in server
        synthdef_name = self.actual_name
        del server._synthdefs[synthdef_name]
        request = supriya.commands.SynthDefFreeRequest(synthdef=self)
        if server.is_running:
            request.communicate(server=server)

    def to_dict(self):
        """
        Convert SynthDef to JSON-serializable dictionay.

        ::

            >>> import json
            >>> result = supriya.assets.synthdefs.default.to_dict()
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "synthdef": {
                    "hash": "da0982184cc8fa54cf9d288a0fe1f6ca",
                    "name": "default",
                    "parameters": {
                        "amplitude": {
                            "range": [
                                0,
                                1
                            ],
                            "rate": "control",
                            "unit": null,
                            "value": 0.1
                        },
                        "frequency": {
                            "range": [
                                0,
                                1
                            ],
                            "rate": "control",
                            "unit": null,
                            "value": 440.0
                        },
                        "gate": {
                            "range": [
                                0,
                                1
                            ],
                            "rate": "control",
                            "unit": null,
                            "value": 1.0
                        },
                        "out": {
                            "range": [
                                0,
                                1
                            ],
                            "rate": "scalar",
                            "unit": null,
                            "value": 0.0
                        },
                        "pan": {
                            "range": [
                                0,
                                1
                            ],
                            "rate": "control",
                            "unit": null,
                            "value": 0.5
                        }
                    }
                }
            }

        """
        result = {
            "name": self.actual_name,
            "hash": self.anonymous_name,
            "parameters": {},
        }
        for parameter_name, parameter in self.parameters.items():
            range_ = [0, 1]
            if parameter.range_:
                range_ = [parameter.range_.minimum, parameter.range_.maximum]
            rate = parameter.parameter_rate.name.lower()
            result["parameters"][parameter_name] = {
                "rate": rate,
                "range": range_,
                "unit": parameter.unit,
                "value": parameter.value,
            }
        result = {"synthdef": result}
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def actual_name(self):
        return self.name or self.anonymous_name

    @property
    def anonymous_name(self):
        md5 = hashlib.md5()
        md5.update(self._compiled_ugen_graph)
        anonymous_name = md5.hexdigest()
        return anonymous_name

    @property
    def audio_channel_count(self):
        return max(self.audio_input_channel_count, self.audio_output_channel_count)

    @property
    def audio_input_channel_count(self):
        """
        Gets audio input channel count of synthdef.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     audio_in = supriya.ugens.In.ar(channel_count=1)
            ...     control_in = supriya.ugens.In.kr(channel_count=2)
            ...     sin = supriya.ugens.SinOsc.ar(
            ...         frequency=audio_in,
            ...     )
            ...     source = audio_in * control_in[1]
            ...     audio_out = supriya.ugens.Out.ar(source=[source] * 4)
            ...
            >>> synthdef = builder.build()

        ::

            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> synthdef.audio_input_channel_count
            1

        Returns integer.
        """
        ugens = tuple(
            _ for _ in self.input_ugens if _.calculation_rate == CalculationRate.AUDIO
        )
        if len(ugens) == 1:
            return ugens[0].channel_count
        elif not ugens:
            return 0
        raise ValueError

    @property
    def audio_output_channel_count(self):
        """
        Gets audio output channel count of synthdef.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     audio_in = supriya.ugens.In.ar(channel_count=1)
            ...     control_in = supriya.ugens.In.kr(channel_count=2)
            ...     sin = supriya.ugens.SinOsc.ar(
            ...         frequency=audio_in,
            ...     )
            ...     source = audio_in * control_in[1]
            ...     audio_out = supriya.ugens.Out.ar(source=[source] * 4)
            ...
            >>> synthdef = builder.build()

        ::

            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> synthdef.audio_output_channel_count
            4

        Returns integer.
        """
        ugens = tuple(
            _ for _ in self.output_ugens if _.calculation_rate == CalculationRate.AUDIO
        )
        if len(ugens) == 1:
            return len(ugens[0].source)
        elif not ugens:
            return 0
        raise ValueError

    @property
    def constants(self):
        return self._constants

    @property
    def control_ugens(self):
        return self._control_ugens

    @property
    def control_channel_count(self):
        return max(self.control_input_channel_count, self.control_output_channel_count)

    @property
    def control_input_channel_count(self):
        """
        Gets control input channel count of synthdef.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     audio_in = supriya.ugens.In.ar(channel_count=1)
            ...     control_in = supriya.ugens.In.kr(channel_count=2)
            ...     sin = supriya.ugens.SinOsc.ar(
            ...         frequency=audio_in,
            ...     )
            ...     source = audio_in * control_in[1]
            ...     audio_out = supriya.ugens.Out.ar(source=[source] * 4)
            ...
            >>> synthdef = builder.build()

        ::

            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> synthdef.control_input_channel_count
            2

        Returns integer.
        """
        ugens = tuple(
            _ for _ in self.input_ugens if _.calculation_rate == CalculationRate.CONTROL
        )
        if len(ugens) == 1:
            return ugens[0].channel_count
        elif not ugens:
            return 0
        raise ValueError

    @property
    def control_output_channel_count(self):
        """
        Gets control output channel count of synthdef.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     audio_in = supriya.ugens.In.ar(channel_count=1)
            ...     control_in = supriya.ugens.In.kr(channel_count=2)
            ...     sin = supriya.ugens.SinOsc.ar(
            ...         frequency=audio_in,
            ...     )
            ...     source = audio_in * control_in[1]
            ...     audio_out = supriya.ugens.Out.ar(source=[source] * 4)
            ...
            >>> synthdef = builder.build()

        ::

            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> synthdef.control_output_channel_count
            0

        Returns integer.
        """
        ugens = tuple(
            _
            for _ in self.output_ugens
            if _.calculation_rate == CalculationRate.CONTROL
        )
        if len(ugens) == 1:
            return len(ugens[0].source)
        elif not ugens:
            return 0
        raise ValueError

    @property
    def done_actions(self):
        done_actions = set()
        for ugen in self.ugens:
            done_action = ugen._get_done_action()
            if done_action is not None:
                done_actions.add(done_action)
        return sorted(done_actions)

    @property
    def has_gate(self):
        return "gate" in self.parameter_names

    @property
    def indexed_parameters(self):
        return self._indexed_parameters

    @property
    def input_ugens(self):
        return tuple(_ for _ in self.ugens if _.is_input_ugen)

    @property
    def is_allocated(self):
        if self.server is not None:
            return self in self.server
        return False

    @property
    def name(self):
        return self._name

    @property
    def output_ugens(self):
        return tuple(_ for _ in self.ugens if _.is_output_ugen)

    @property
    def parameters(self):
        return {
            parameter.name: parameter for index, parameter in self.indexed_parameters
        }

    @property
    def parameter_names(self):
        return [parameter.name for index, parameter in self.indexed_parameters]

    @property
    def ugens(self):
        return self._ugens


class UGenSortBundle(SupriyaObject):

    ### INITIALIZER ###

    def __init__(self, ugen, width_first_antecedents):
        self.antecedents = []
        self.descendants = []
        self.ugen = ugen
        self.width_first_antecedents = tuple(width_first_antecedents)

    ### PRIVATE METHODS ###

    def _initialize_topological_sort(self, sort_bundles):
        for input_ in self.ugen.inputs:
            if isinstance(input_, OutputProxy):
                input_ = input_.source
            elif not isinstance(input_, UGen):
                continue
            input_sort_bundle = sort_bundles[input_]
            if input_ not in self.antecedents:
                self.antecedents.append(input_)
            if self.ugen not in input_sort_bundle.descendants:
                input_sort_bundle.descendants.append(self.ugen)
        for input_ in self.width_first_antecedents:
            input_sort_bundle = sort_bundles[input_]
            if input_ not in self.antecedents:
                self.antecedents.append(input_)
            if self.ugen not in input_sort_bundle.descendants:
                input_sort_bundle.descendants.append(self.ugen)

    def _make_available(self, available_ugens):
        if not self.antecedents:
            if self.ugen not in available_ugens:
                available_ugens.append(self.ugen)

    def _schedule(self, available_ugens, out_stack, sort_bundles):
        for ugen in reversed(self.descendants):
            sort_bundle = sort_bundles[ugen]
            sort_bundle.antecedents.remove(self.ugen)
            sort_bundle._make_available(available_ugens)
        out_stack.append(self.ugen)

    ### PUBLIC METHODS ###

    def clear(self):
        self.antecedents[:] = []
        self.descendants[:] = []
        self.width_first_antecedents[:] = []


class SuperColliderSynthDef(SupriyaObject):

    ### CLASS VARIABLES ###

    __slots__ = ("_body", "_name", "_rates")

    ### INITIALIZER ###

    def __init__(self, name, body, rates=None):
        self._name = name
        self._body = body
        self._rates = rates

    ### PRIVATE METHODS ###

    def _build_sc_input(self, directory_path):
        input_ = []
        input_.append("a = SynthDef(")
        input_.append("    \\{}, {{".format(self.name))
        for line in self.body.splitlines():
            input_.append("    " + line)
        if self.rates:
            input_.append("}}, {});".format(self.rates))
        else:
            input_.append("});")
        input_.append('"Defined SynthDef".postln;')
        input_.append('a.writeDefFile("{}");'.format(directory_path))
        input_.append('"Wrote SynthDef".postln;')
        input_.append("0.exit;")
        input_ = "\n".join(input_)
        return input_

    ### PUBLIC METHODS ###

    def compile(self):
        sclang_path = sclang.find()
        with tempfile.TemporaryDirectory() as directory:
            directory_path = pathlib.Path(directory)
            sc_input = self._build_sc_input(directory_path)
            print(sc_input)
            sc_file_path = directory_path / f"{self.name}.sc"
            sc_file_path.write_text(sc_input)
            command = " ".join([str(sclang_path), "-D", str(sc_file_path)])
            print(command)
            subprocess.run(command, shell=True)
            result = (directory_path / f"{self.name}.scsyndef").read_bytes()
        return bytes(result)

    ### PUBLIC PROPERTIES ###

    @property
    def body(self):
        return self._body

    @property
    def rates(self):
        return self._rates

    @property
    def name(self):
        return self._name



from supriya import AddAction
from supriya.enums import RequestId

from .bases import Request, Response


class SynthNewRequest(Request):
    """
    A /s_new request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.SynthNewRequest(
        ...     add_action=supriya.AddAction.ADD_TO_TAIL,
        ...     node_id=1001,
        ...     synthdef="test",
        ...     target_node_id=1000,
        ...     frequency=443,
        ...     phase=0.2,
        ... )
        >>> request
        SynthNewRequest(
            add_action=AddAction.ADD_TO_TAIL,
            frequency=443,
            node_id=1001,
            phase=0.2,
            synthdef='test',
            target_node_id=1000,
        )

    ::

        >>> request.to_osc()
        OscMessage('/s_new', 'test', 1001, 1, 1000, 'frequency', 443, 'phase', 0.2)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTH_NEW

    ### INITIALIZER ###

    def __init__(
        self,
        add_action=None,
        node_id=None,
        synthdef=None,
        target_node_id=None,
        **kwargs,
    ):
        import supriya.synthdefs

        Request.__init__(self)
        self._add_action = AddAction.from_expr(add_action)
        self._node_id = node_id
        prototype = (str, supriya.synthdefs.SynthDef)
        assert isinstance(synthdef, prototype)
        self._synthdef = synthdef
        self._target_node_id = target_node_id
        self._kwargs = tuple(sorted(kwargs.items()))

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        from supriya.realtime import Node, Synth

        if isinstance(self.node_id, Synth):
            node_id = None
            synth = self.node_id
        else:
            node_id = self.node_id
            synth = Synth(synthdef=self.synthdef, **dict(self.kwargs))
        if isinstance(self.target_node_id, Node):
            target_node = self.target_node_id
        else:
            target_node = server._nodes[self.target_node_id]
        synth._register_with_local_server(
            server, node_id=node_id, node_id_is_permanent=synth.node_id_is_permanent
        )
        target_node._move_node(add_action=self.add_action, node=synth)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        import supriya.synthdefs

        request_id = self.request_name
        synthdef = self.synthdef
        if isinstance(synthdef, supriya.synthdefs.SynthDef):
            synthdef = synthdef.actual_name
        node_id = self._sanitize_node_id(self.node_id, with_placeholders)
        add_action = int(self.add_action)
        target_node_id = self._sanitize_node_id(self.target_node_id, with_placeholders)
        if not isinstance(target_node_id, int) and with_placeholders:
            target_node_id = -1
        contents = [request_id, synthdef, node_id, add_action, target_node_id]
        for key, value in self._kwargs:
            contents.append(key)
            contents.append(value)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def add_action(self):
        return self._add_action

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def node_id(self):
        return self._node_id

    @property
    def response_patterns(self):
        return ["/n_go", int(self.node_id)], None

    @property
    def synthdef(self):
        return self._synthdef

    @property
    def target_node_id(self):
        return self._target_node_id


class TriggerResponse(Response):

    ### INITIALIZER ###

    def __init__(self, node_id=None, trigger_id=None, trigger_value=None):
        self._node_id = node_id
        self._trigger_id = trigger_id
        self._trigger_value = trigger_value

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        arguments = osc_message.contents
        response = cls(*arguments)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id

    @property
    def trigger_id(self):
        return self._trigger_id

    @property
    def trigger_value(self):
        return self._trigger_value



import typing
from collections.abc import Sequence

import supriya.osc
from supriya.enums import RequestId
from supriya.querytree import QueryTreeControl, QueryTreeGroup, QueryTreeSynth

from .bases import Request, Response


class GroupDeepFreeRequest(Request):
    """
    A /g_deepFree request.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)
        >>> group.extend([supriya.Synth(), supriya.Group()])
        >>> group[1].extend([supriya.Synth(), supriya.Group()])
        >>> group[1][1].extend([supriya.Synth(), supriya.Group()])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1002 group
                        1003 default
                            out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                        1004 group
                            1005 default
                                out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                            1006 group

    ::

        >>> request = supriya.commands.GroupDeepFreeRequest(group)
        >>> request.to_osc()
        OscMessage('/g_deepFree', 1000)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/g_deepFree', 1000))
        ('S', OscMessage('/sync', 2))
        ('R', OscMessage('/n_end', 1001, -1, -1, -1, 0))
        ('R', OscMessage('/n_end', 1003, -1, -1, -1, 0))
        ('R', OscMessage('/n_end', 1005, -1, -1, -1, 0))
        ('R', OscMessage('/synced', 2))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1002 group
                        1004 group
                            1006 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                    1002 group
                        1004 group
                            1006 group

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.GROUP_DEEP_FREE

    ### INITIALIZER ###

    def __init__(self, group_id):
        Request.__init__(self)
        self._group_id = group_id

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        group_id = int(self.group_id)
        message = supriya.osc.OscMessage(request_id, group_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def group_id(self):
        return self._group_id


class GroupFreeAllRequest(Request):
    """
    A /g_freeAll request.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)
        >>> group.extend([supriya.Synth(), supriya.Group()])
        >>> group[1].extend([supriya.Synth(), supriya.Group()])
        >>> group[1][1].extend([supriya.Synth(), supriya.Group()])

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1001 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                    1002 group
                        1003 default
                            out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                        1004 group
                            1005 default
                                out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5
                            1006 group

    ::

        >>> request = supriya.commands.GroupFreeAllRequest(group)
        >>> request.to_osc()
        OscMessage('/g_freeAll', 1000)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/g_freeAll', 1000))
        ('S', OscMessage('/sync', 2))
        ('R', OscMessage('/n_end', 1001, -1, -1, -1, 0))
        ('R', OscMessage('/n_end', 1003, -1, -1, -1, 0))
        ('R', OscMessage('/n_end', 1005, -1, -1, -1, 0))
        ('R', OscMessage('/n_end', 1006, -1, -1, -1, 1, -1, -1))
        ('R', OscMessage('/n_end', 1004, -1, -1, -1, 1, -1, -1))
        ('R', OscMessage('/n_end', 1002, -1, -1, -1, 1, -1, -1))
        ('R', OscMessage('/synced', 2))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.GROUP_FREE_ALL

    ### INITIALIZER ###

    def __init__(self, group_id):
        Request.__init__(self)
        self._group_id = group_id

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        group_id = int(self.group_id)
        message = supriya.osc.OscMessage(request_id, group_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def group_id(self):
        return self._group_id


class GroupNewRequest(Request):
    """
    A /g_new request.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group

    ::

        >>> request = supriya.commands.GroupNewRequest(
        ...     items=[
        ...         supriya.commands.GroupNewRequest.Item(
        ...             add_action=supriya.AddAction.ADD_TO_TAIL,
        ...             node_id=1001,
        ...             target_node_id=1,
        ...         ),
        ...         supriya.commands.GroupNewRequest.Item(
        ...             add_action=supriya.AddAction.ADD_TO_HEAD,
        ...             node_id=1002,
        ...             target_node_id=1001,
        ...         ),
        ...     ],
        ... )
        >>> request.to_osc()
        OscMessage('/g_new', 1001, 1, 1, 1002, 0, 1001)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     _ = request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/g_new', 1001, 1, 1, 1002, 0, 1001))
        ('R', OscMessage('/n_go', 1001, 1, 1000, -1, 1, -1, -1))
        ('R', OscMessage('/n_go', 1002, 1001, -1, -1, 1, -1, -1))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                1001 group
                    1002 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                1001 group
                    1002 group

    """

    ### CLASS VARIABLES ###

    class Item(typing.NamedTuple):
        add_action: int = 0
        node_id: int = 0
        target_node_id: int = 0

    request_id = RequestId.GROUP_NEW

    ### INITIALIZER ###

    def __init__(self, items=None):
        # TODO: Support multi-group allocation
        Request.__init__(self)
        if items:
            if not isinstance(items, Sequence):
                items = [items]
            items = list(items)
            for i, (add_action, node_id, target_node_id) in enumerate(items):
                add_action = supriya.AddAction.from_expr(add_action)
                items[i] = self.Item(
                    add_action=add_action,
                    node_id=node_id,
                    target_node_id=target_node_id,
                )
        self._items = items

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        from supriya.realtime import Group, Node

        for item in self.items:
            if isinstance(item.node_id, Group):
                node_id = None
                group = item.node_id
            else:
                node_id = item.node_id
                group = Group()
            if isinstance(item.target_node_id, Node):
                target_node = item.target_node_id
            else:
                target_node = server._nodes[item.target_node_id]
            group._register_with_local_server(
                server, node_id=node_id, node_id_is_permanent=group.node_id_is_permanent
            )
            target_node._move_node(add_action=item.add_action, node=group)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        for item in self.items:
            node_id = self._sanitize_node_id(item.node_id, with_placeholders)
            target_node_id = self._sanitize_node_id(
                item.target_node_id, with_placeholders
            )
            contents.append(node_id)
            contents.append(int(item.add_action))
            contents.append(target_node_id)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        return self._items

    @property
    def response_patterns(self):
        return ["/n_go", int(self.items[-1].node_id)], None


class GroupQueryTreeRequest(Request):
    """
    A /g_queryTree request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.GroupQueryTreeRequest(
        ...     node_id=0,
        ...     include_controls=True,
        ... )
        >>> request
        GroupQueryTreeRequest(
            include_controls=True,
            node_id=0,
        )

    ::

        >>> request.to_osc()
        OscMessage('/g_queryTree', 0, 1)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.GROUP_QUERY_TREE

    ### INITIALIZER ###

    def __init__(self, node_id=None, include_controls=False):
        Request.__init__(self)
        self._node_id = node_id
        self._include_controls = bool(include_controls)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        node_id = int(self.node_id)
        include_controls = int(self.include_controls)
        message = supriya.osc.OscMessage(request_id, node_id, include_controls)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def include_controls(self):
        return self._include_controls

    @property
    def node_id(self):
        return self._node_id

    @property
    def response_patterns(self):
        return ["/g_queryTree.reply", int(self.include_controls), self.node_id], None


class ParallelGroupNewRequest(GroupNewRequest):
    """
    A /p_new request.

    ..  note::  This behaves like a ``/g_new`` request when run on ``scsynth``
                instead of ``supernova``.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group

    ::

        >>> request = supriya.commands.ParallelGroupNewRequest(
        ...     items=[
        ...         supriya.commands.ParallelGroupNewRequest.Item(
        ...             add_action=supriya.AddAction.ADD_TO_TAIL,
        ...             node_id=1001,
        ...             target_node_id=1,
        ...         ),
        ...     ],
        ... )
        >>> request.to_osc()
        OscMessage('/p_new', 1001, 1, 1)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     response = request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/p_new', 1001, 1, 1))
        ('R', OscMessage('/n_go', 1001, 1, 1000, -1, 1, -1, -1))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                1001 group

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                1001 group

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.PARALLEL_GROUP_NEW


class QueryTreeResponse(Response):

    ### INITIALIZER ###

    def __init__(self, node_id=None, query_tree_group=None):
        self._node_id = node_id
        self._query_tree_group = query_tree_group

    ### SPECIAL METHODS ###

    def __str__(self):
        return str(self._query_tree_group)

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        """
        Create response from OSC message.

        ::

            >>> message = supriya.osc.OscMessage(
            ...     "/g_queryTree.reply", 0, 0, 1, 1, 2, 1001, 0, 1000, 1, 1002, 0
            ... )
            >>> supriya.commands.QueryTreeResponse.from_osc_message(message)
            QueryTreeResponse(
                node_id=0,
                query_tree_group=QueryTreeGroup(
                    children=(
                        QueryTreeGroup(
                            children=(
                                QueryTreeGroup(
                                    children=(),
                                    node_id=1001,
                                ),
                                QueryTreeGroup(
                                    children=(
                                        QueryTreeGroup(
                                            children=(),
                                            node_id=1002,
                                        ),
                                    ),
                                    node_id=1000,
                                ),
                            ),
                            node_id=1,
                        ),
                    ),
                    node_id=0,
                ),
            )

        ::

            >>> print(supriya.commands.QueryTreeResponse.from_osc_message(message))
            NODE TREE 0 group
                1 group
                    1001 group
                    1000 group
                        1002 group

        """

        def recurse(contents, control_flag):
            node_id = contents.pop(0)
            child_count = contents.pop(0)
            if child_count == -1:
                controls = []
                synthdef_name = contents.pop(0)
                if control_flag:
                    control_count = contents.pop(0)
                    for i in range(control_count):
                        control_name_or_index = contents.pop(0)
                        control_value = contents.pop(0)
                        control = QueryTreeControl(
                            control_name_or_index=control_name_or_index,
                            control_value=control_value,
                        )
                        controls.append(control)
                controls = tuple(controls)
                result = QueryTreeSynth(
                    node_id=node_id, synthdef_name=synthdef_name, controls=controls
                )
            else:
                children = []
                for i in range(child_count):
                    children.append(recurse(contents, control_flag))
                children = tuple(children)
                result = QueryTreeGroup(node_id=node_id, children=children)
            return result

        contents = list(osc_message.contents)
        control_flag = bool(contents.pop(0))
        query_tree_group = recurse(contents, control_flag)
        response = cls(
            node_id=query_tree_group.node_id, query_tree_group=query_tree_group
        )
        return response

    def to_dict(self, flat=False):
        """
        Convert QueryTreeResponse to JSON-serializable dictionary.

        ::

            >>> query_tree_response = supriya.commands.QueryTreeResponse(
            ...     node_id=0,
            ...     query_tree_group=supriya.querytree.QueryTreeGroup(
            ...         node_id=0,
            ...         children=(
            ...             supriya.querytree.QueryTreeGroup(
            ...                 node_id=1,
            ...                 children=(
            ...                     supriya.querytree.QueryTreeGroup(
            ...                         node_id=1002,
            ...                         children=(
            ...                             supriya.querytree.QueryTreeSynth(
            ...                                 node_id=1105,
            ...                                 synthdef_name="dca557070c6b57164557041ac746fb3f",
            ...                                 controls=(
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="damping",
            ...                                         control_value=0.06623425334692,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="duration",
            ...                                         control_value=3.652155876159668,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="level",
            ...                                         control_value=0.894767701625824,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="out",
            ...                                         control_value=16.0,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="room_size",
            ...                                         control_value=0.918643176555634,
            ...                                     ),
            ...                                 ),
            ...                             ),
            ...                             supriya.querytree.QueryTreeSynth(
            ...                                 node_id=1098,
            ...                                 synthdef_name="cc754c63533fdcf412a44ef6adb1a8f0",
            ...                                 controls=(
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="duration",
            ...                                         control_value=5.701356887817383,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="level",
            ...                                         control_value=0.959683060646057,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="out",
            ...                                         control_value=16.0,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="pitch_dispersion",
            ...                                         control_value=0.040342573076487,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="pitch_shift",
            ...                                         control_value=10.517594337463379,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="time_dispersion",
            ...                                         control_value=0.666014134883881,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="window_size",
            ...                                         control_value=1.014111995697021,
            ...                                     ),
            ...                                 ),
            ...                             ),
            ...                             supriya.querytree.QueryTreeSynth(
            ...                                 node_id=1096,
            ...                                 synthdef_name="5cb6fb104ee1dc44d6b300e13112d37a",
            ...                                 controls=(
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="duration",
            ...                                         control_value=5.892660140991211,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="level",
            ...                                         control_value=0.159362614154816,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="out",
            ...                                         control_value=16.0,
            ...                                     ),
            ...                                 ),
            ...                             ),
            ...                             supriya.querytree.QueryTreeSynth(
            ...                                 node_id=1010,
            ...                                 synthdef_name="da0982184cc8fa54cf9d288a0fe1f6ca",
            ...                                 controls=(
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="out",
            ...                                         control_value=16.0,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="amplitude",
            ...                                         control_value=0.846831738948822,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="frequency",
            ...                                         control_value=1522.9603271484375,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="gate",
            ...                                         control_value=0.0,
            ...                                     ),
            ...                                     supriya.querytree.QueryTreeControl(
            ...                                         control_name_or_index="pan",
            ...                                         control_value=0.733410477638245,
            ...                                     ),
            ...                                 ),
            ...                             ),
            ...                         ),
            ...                     ),
            ...                     supriya.querytree.QueryTreeSynth(
            ...                         node_id=1003,
            ...                         synthdef_name="454b69a7c505ddecc5b39762d291a5ec",
            ...                         controls=(
            ...                             supriya.querytree.QueryTreeControl(
            ...                                 control_name_or_index="done_action",
            ...                                 control_value=2.0,
            ...                             ),
            ...                             supriya.querytree.QueryTreeControl(
            ...                                 control_name_or_index="fade_time",
            ...                                 control_value=0.019999999552965,
            ...                             ),
            ...                             supriya.querytree.QueryTreeControl(
            ...                                 control_name_or_index="gate",
            ...                                 control_value=1.0,
            ...                             ),
            ...                             supriya.querytree.QueryTreeControl(
            ...                                 control_name_or_index="in_",
            ...                                 control_value=16.0,
            ...                             ),
            ...                             supriya.querytree.QueryTreeControl(
            ...                                 control_name_or_index="out",
            ...                                 control_value=0.0,
            ...                             ),
            ...                         ),
            ...                     ),
            ...                 ),
            ...             ),
            ...             supriya.querytree.QueryTreeSynth(
            ...                 node_id=1000,
            ...                 synthdef_name="72696657e1216698c415e704ea8ab9a2",
            ...                 controls=(
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_clamp_time",
            ...                         control_value=0.009999999776483,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_postgain",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_pregain",
            ...                         control_value=6.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_relax_time",
            ...                         control_value=0.100000001490116,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_slope_above",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_slope_below",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_1_threshold",
            ...                         control_value=0.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_clamp_time",
            ...                         control_value=0.009999999776483,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_postgain",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_pregain",
            ...                         control_value=3.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_relax_time",
            ...                         control_value=0.100000001490116,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_slope_above",
            ...                         control_value=0.5,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_slope_below",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_2_threshold",
            ...                         control_value=-6.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_clamp_time",
            ...                         control_value=0.009999999776483,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_postgain",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_pregain",
            ...                         control_value=-3.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_relax_time",
            ...                         control_value=0.100000001490116,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_slope_above",
            ...                         control_value=0.25,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_slope_below",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_3_threshold",
            ...                         control_value=-12.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_clamp_time",
            ...                         control_value=0.009999999776483,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_postgain",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_pregain",
            ...                         control_value=-3.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_relax_time",
            ...                         control_value=0.100000001490116,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_slope_above",
            ...                         control_value=0.125,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_slope_below",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="band_4_threshold",
            ...                         control_value=-18.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="frequency_1",
            ...                         control_value=200.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="frequency_2",
            ...                         control_value=2000.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="frequency_3",
            ...                         control_value=5000.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="in_",
            ...                         control_value=0.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="out",
            ...                         control_value=0.0,
            ...                     ),
            ...                 ),
            ...             ),
            ...             supriya.querytree.QueryTreeSynth(
            ...                 node_id=1001,
            ...                 synthdef_name="c1aa521afab5b0c0ce3d744690951649",
            ...                 controls=(
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="level",
            ...                         control_value=1.0,
            ...                     ),
            ...                     supriya.querytree.QueryTreeControl(
            ...                         control_name_or_index="out",
            ...                         control_value=0.0,
            ...                     ),
            ...                 ),
            ...             ),
            ...         ),
            ...     ),
            ... )

        ::

            >>> import json
            >>> result = query_tree_response.to_dict()
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "server_tree": {
                    "children": [
                        {
                            "children": [
                                {
                                    "children": [
                                        {
                                            "controls": {
                                                "damping": 0.06623425334692,
                                                "duration": 3.652155876159668,
                                                "level": 0.894767701625824,
                                                "out": 16.0,
                                                "room_size": 0.918643176555634
                                            },
                                            "node_id": 1105,
                                            "synthdef": "dca557070c6b57164557041ac746fb3f"
                                        },
                                        {
                                            "controls": {
                                                "duration": 5.701356887817383,
                                                "level": 0.959683060646057,
                                                "out": 16.0,
                                                "pitch_dispersion": 0.040342573076487,
                                                "pitch_shift": 10.517594337463379,
                                                "time_dispersion": 0.666014134883881,
                                                "window_size": 1.014111995697021
                                            },
                                            "node_id": 1098,
                                            "synthdef": "cc754c63533fdcf412a44ef6adb1a8f0"
                                        },
                                        {
                                            "controls": {
                                                "duration": 5.892660140991211,
                                                "level": 0.159362614154816,
                                                "out": 16.0
                                            },
                                            "node_id": 1096,
                                            "synthdef": "5cb6fb104ee1dc44d6b300e13112d37a"
                                        },
                                        {
                                            "controls": {
                                                "amplitude": 0.846831738948822,
                                                "frequency": 1522.9603271484375,
                                                "gate": 0.0,
                                                "out": 16.0,
                                                "pan": 0.733410477638245
                                            },
                                            "node_id": 1010,
                                            "synthdef": "da0982184cc8fa54cf9d288a0fe1f6ca"
                                        }
                                    ],
                                    "node_id": 1002
                                },
                                {
                                    "controls": {
                                        "done_action": 2.0,
                                        "fade_time": 0.019999999552965,
                                        "gate": 1.0,
                                        "in_": 16.0,
                                        "out": 0.0
                                    },
                                    "node_id": 1003,
                                    "synthdef": "454b69a7c505ddecc5b39762d291a5ec"
                                }
                            ],
                            "node_id": 1
                        },
                        {
                            "controls": {
                                "band_1_clamp_time": 0.009999999776483,
                                "band_1_postgain": 1.0,
                                "band_1_pregain": 6.0,
                                "band_1_relax_time": 0.100000001490116,
                                "band_1_slope_above": 1.0,
                                "band_1_slope_below": 1.0,
                                "band_1_threshold": 0.0,
                                "band_2_clamp_time": 0.009999999776483,
                                "band_2_postgain": 1.0,
                                "band_2_pregain": 3.0,
                                "band_2_relax_time": 0.100000001490116,
                                "band_2_slope_above": 0.5,
                                "band_2_slope_below": 1.0,
                                "band_2_threshold": -6.0,
                                "band_3_clamp_time": 0.009999999776483,
                                "band_3_postgain": 1.0,
                                "band_3_pregain": -3.0,
                                "band_3_relax_time": 0.100000001490116,
                                "band_3_slope_above": 0.25,
                                "band_3_slope_below": 1.0,
                                "band_3_threshold": -12.0,
                                "band_4_clamp_time": 0.009999999776483,
                                "band_4_postgain": 1.0,
                                "band_4_pregain": -3.0,
                                "band_4_relax_time": 0.100000001490116,
                                "band_4_slope_above": 0.125,
                                "band_4_slope_below": 1.0,
                                "band_4_threshold": -18.0,
                                "frequency_1": 200.0,
                                "frequency_2": 2000.0,
                                "frequency_3": 5000.0,
                                "in_": 0.0,
                                "out": 0.0
                            },
                            "node_id": 1000,
                            "synthdef": "72696657e1216698c415e704ea8ab9a2"
                        },
                        {
                            "controls": {
                                "level": 1.0,
                                "out": 0.0
                            },
                            "node_id": 1001,
                            "synthdef": "c1aa521afab5b0c0ce3d744690951649"
                        }
                    ],
                    "node_id": 0
                }
            }

        ::

            >>> result = query_tree_response.to_dict(flat=True)
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "server_tree": [
                    {
                        "children": [
                            1,
                            1000,
                            1001
                        ],
                        "node_id": 0,
                        "parent": null
                    },
                    {
                        "children": [
                            1002,
                            1003
                        ],
                        "node_id": 1,
                        "parent": 0
                    },
                    {
                        "children": [
                            1105,
                            1098,
                            1096,
                            1010
                        ],
                        "node_id": 1002,
                        "parent": 1
                    },
                    {
                        "controls": {
                            "damping": 0.06623425334692,
                            "duration": 3.652155876159668,
                            "level": 0.894767701625824,
                            "out": 16.0,
                            "room_size": 0.918643176555634
                        },
                        "node_id": 1105,
                        "parent": 1002,
                        "synthdef": "dca557070c6b57164557041ac746fb3f"
                    },
                    {
                        "controls": {
                            "duration": 5.701356887817383,
                            "level": 0.959683060646057,
                            "out": 16.0,
                            "pitch_dispersion": 0.040342573076487,
                            "pitch_shift": 10.517594337463379,
                            "time_dispersion": 0.666014134883881,
                            "window_size": 1.014111995697021
                        },
                        "node_id": 1098,
                        "parent": 1002,
                        "synthdef": "cc754c63533fdcf412a44ef6adb1a8f0"
                    },
                    {
                        "controls": {
                            "duration": 5.892660140991211,
                            "level": 0.159362614154816,
                            "out": 16.0
                        },
                        "node_id": 1096,
                        "parent": 1002,
                        "synthdef": "5cb6fb104ee1dc44d6b300e13112d37a"
                    },
                    {
                        "controls": {
                            "amplitude": 0.846831738948822,
                            "frequency": 1522.9603271484375,
                            "gate": 0.0,
                            "out": 16.0,
                            "pan": 0.733410477638245
                        },
                        "node_id": 1010,
                        "parent": 1002,
                        "synthdef": "da0982184cc8fa54cf9d288a0fe1f6ca"
                    },
                    {
                        "controls": {
                            "done_action": 2.0,
                            "fade_time": 0.019999999552965,
                            "gate": 1.0,
                            "in_": 16.0,
                            "out": 0.0
                        },
                        "node_id": 1003,
                        "parent": 1,
                        "synthdef": "454b69a7c505ddecc5b39762d291a5ec"
                    },
                    {
                        "controls": {
                            "band_1_clamp_time": 0.009999999776483,
                            "band_1_postgain": 1.0,
                            "band_1_pregain": 6.0,
                            "band_1_relax_time": 0.100000001490116,
                            "band_1_slope_above": 1.0,
                            "band_1_slope_below": 1.0,
                            "band_1_threshold": 0.0,
                            "band_2_clamp_time": 0.009999999776483,
                            "band_2_postgain": 1.0,
                            "band_2_pregain": 3.0,
                            "band_2_relax_time": 0.100000001490116,
                            "band_2_slope_above": 0.5,
                            "band_2_slope_below": 1.0,
                            "band_2_threshold": -6.0,
                            "band_3_clamp_time": 0.009999999776483,
                            "band_3_postgain": 1.0,
                            "band_3_pregain": -3.0,
                            "band_3_relax_time": 0.100000001490116,
                            "band_3_slope_above": 0.25,
                            "band_3_slope_below": 1.0,
                            "band_3_threshold": -12.0,
                            "band_4_clamp_time": 0.009999999776483,
                            "band_4_postgain": 1.0,
                            "band_4_pregain": -3.0,
                            "band_4_relax_time": 0.100000001490116,
                            "band_4_slope_above": 0.125,
                            "band_4_slope_below": 1.0,
                            "band_4_threshold": -18.0,
                            "frequency_1": 200.0,
                            "frequency_2": 2000.0,
                            "frequency_3": 5000.0,
                            "in_": 0.0,
                            "out": 0.0
                        },
                        "node_id": 1000,
                        "parent": 0,
                        "synthdef": "72696657e1216698c415e704ea8ab9a2"
                    },
                    {
                        "controls": {
                            "level": 1.0,
                            "out": 0.0
                        },
                        "node_id": 1001,
                        "parent": 0,
                        "synthdef": "c1aa521afab5b0c0ce3d744690951649"
                    }
                ]
            }

        """

        def recurse(node, parent_node_id, nodes):
            if "synthdef" in node:
                node["parent"] = parent_node_id
                nodes.append(node)
            else:
                group = {
                    "node_id": node["node_id"],
                    "parent": parent_node_id,
                    "children": [x["node_id"] for x in node["children"]],
                }
                nodes.append(group)
                for x in node["children"]:
                    recurse(x, node["node_id"], nodes)

        data = self.query_tree_group.to_dict()
        if not flat:
            return {"server_tree": data}
        nodes = []
        recurse(data, None, nodes)
        return {"server_tree": nodes}

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id

    @property
    def query_tree_group(self):
        return self._query_tree_group
